"""Sports Reference game model."""

# pylint: disable=too-many-locals,too-many-statements,unused-argument,protected-access,too-many-arguments,use-maxsplit-arg,too-many-branches,duplicate-code
import datetime
import io
import logging
import re
import urllib.parse

import dateutil
import pandas as pd
import pytest_is_running
import requests
from bs4 import BeautifulSoup, Tag
from dateutil.parser import parse

from ...cache import MEMORY
from ...proxy_session import ProxySession
from ..game_model import GameModel
from ..league import League
from ..season_type import SeasonType
from ..team_model import TeamModel
from .sportsreference_team_model import create_sportsreference_team_model
from .sportsreference_venue_model import create_sportsreference_venue_model

_MONTHS = [
    "January",
    "February",
    "March",
    "April",
    "May",
    "June",
    "July",
    "August",
    "September",
    "October",
    "November",
    "December",
]
_NUMBER_PARENTHESIS_PATTERN = r"\(\d+\)"


def _find_old_dt(
    dfs: list[pd.DataFrame],
    session: ProxySession,
    soup: BeautifulSoup,
    url: str,
    league: League,
    player_urls: set[str],
    fg: dict[str, int],
    fga: dict[str, int],
    offensive_rebounds: dict[str, int],
    assists: dict[str, int],
    turnovers: dict[str, int],
    response: requests.Response,
    positions_validator: dict[str, str],
) -> tuple[datetime.datetime, list[TeamModel], str | None]:
    teams: list[TeamModel] = []

    def _process_team_row(df: pd.DataFrame):
        team_rows = [str(df.iat[0, x]) for x in range(len(df.columns.values))]
        team_rows = [x for x in team_rows if x != "nan"]
        if len(team_rows) == 1:
            team_row = team_rows[0]
            for sentinel in ["Next Game", "Prev Game"]:
                if sentinel in team_row:
                    team_rows = [
                        team_row[: team_row.find(sentinel) + len(sentinel)].strip(),
                        team_row[team_row.find(sentinel) + len(sentinel) :]
                        .strip()
                        .replace("/ Next Game ⇒", "")
                        .replace("⇒", "")
                        .strip(),
                    ]
                    last_splits = team_rows[1].split()
                    if last_splits:
                        if last_splits[0] in _MONTHS:
                            continue
                    break
            if not team_rows[1]:
                sentinel = "Next Game"
                team_rows = [
                    team_row[: team_row.find(sentinel) + len(sentinel)].strip(),
                    team_row[team_row.find(sentinel) + len(sentinel) :]
                    .strip()
                    .replace("⇒", "")
                    .strip(),
                ]
            if not team_rows[1]:
                sentinel = "Prev Game"
                team_rows = [
                    team_row[: team_row.find(sentinel) + len(sentinel)].strip(),
                    team_row[team_row.find(sentinel) + len(sentinel) :]
                    .strip()
                    .replace("⇒", "")
                    .strip(),
                ]
        for team_row in team_rows:
            if team_row.startswith("⇒"):
                team_row = team_row[1:]
            team_row = team_row.strip()
            if "Prev Game" in team_row:
                team_name_points = (
                    team_row.split("⇐")[0]
                    .strip()
                    .replace("Prev Game", "")
                    # .replace("/", "")
                    .strip()
                )
            else:
                team_name_points = (
                    team_row.split("⇒")[0].replace("Next Game", "").strip()
                )
            # Handle team name points like "Indiana Pacers 97 45-37 (Won 3)"
            if "-" in team_name_points:
                dash_splits = team_name_points.split("-")
                # Handle team name points like "Kansas City-Omaha Kings 89"
                # and cases like "Kansas City-Omaha Kings 95 44-38 (Won 1) ⇐ Prev Game"
                for i in range(len(dash_splits) - 1):
                    if (
                        dash_splits[i].split()[-1].isdigit()
                        and dash_splits[i + 1].split()[0].isdigit()
                    ):
                        team_name_points = " ".join(
                            "-".join(dash_splits[: i + 1]).strip().split()[:-1]
                        )
                        break
            for marker in ["Lost", "Won"]:
                team_name_points = team_name_points.split(marker)[0].strip()
            points = int(team_name_points.split()[-1].strip())
            team_name = " ".join(team_name_points.split()[:-1]).strip()
            if " at " in team_name:
                team_name = team_name.split(" at ")[0].strip()
            if " vs " in team_name:
                team_name = team_name.split(" vs ")[0].strip()
            for month in _MONTHS:
                month_split = " " + month + " "
                if month_split in team_name:
                    team_name = team_name.split(month_split)[0].strip()
                    points = int(team_name.split()[-1].strip())
                    team_name = " ".join(team_name.split()[:-1]).strip()
                    break
            team_name = " ".join(
                re.sub(_NUMBER_PARENTHESIS_PATTERN, "", team_name).split()
            ).strip()
            team_a = soup.find("a", text=team_name, href=True)
            if not isinstance(team_a, Tag):
                logging.error(team_name)
                logging.error(response.url)
                logging.error(response.text)
                raise ValueError("team_a is not a tag.")
            team_url = urllib.parse.urljoin(url, str(team_a.get("href")))
            if dt is None:
                raise ValueError("dt is null.")
            teams.append(
                create_sportsreference_team_model(
                    session,
                    team_url,
                    dt,
                    league,
                    player_urls,
                    points,
                    fg,
                    fga,
                    offensive_rebounds,
                    assists,
                    turnovers,
                    team_name,
                    positions_validator,
                )
            )

    dt = None
    venue_name = None
    for df in dfs:
        if len(df) == 2:
            test_row = df.iat[0, 0]
            test_row_2 = df.iat[1, 0]
            if isinstance(test_row, float):
                continue

            try:
                if (
                    "Prev Game" in test_row
                    or "Next Game" in test_row
                    or "Lost" in test_row
                    or "PM," in test_row_2
                ):
                    date_venue_split = df.iat[1, 0].split()
                    current_idx = 5
                    try:
                        dt = parse(" ".join(date_venue_split[:current_idx]))
                    except dateutil.parser._parser.ParserError:  # type: ignore
                        try:
                            current_idx = 3
                            dt = parse(" ".join(date_venue_split[:current_idx]))
                        except dateutil.parser._parser.ParserError:  # type: ignore
                            dt = parse(
                                " ".join(",".join(test_row.split(",")[1:]).split()[:3])
                            )
                    venue_name = " ".join(date_venue_split[current_idx:])
                    _process_team_row(df)
                    break
            except TypeError as exc:
                logging.error(test_row)
                logging.error(response.url)
                logging.error(response.text)
                raise exc

    if dt is None:
        title_tag = soup.find("title")
        if not isinstance(title_tag, Tag):
            raise ValueError("title_tag is not a tag.")
        title = title_tag.get_text().strip().split("|")[0].strip()
        date = title[title.find(",") :].strip()
        dt = parse(date)
        for df in dfs:
            test_row = df.iat[0, 0]
            if isinstance(test_row, float):
                continue
            try:
                if "Prev Game" in test_row:
                    _process_team_row(df)
                    break
            except TypeError as exc:
                logging.error(test_row)
                logging.error(response.url)
                logging.error(response.text)
                raise exc

    if venue_name is not None and not venue_name.replace(",", "").strip():
        venue_name = None

    if dt is None:
        raise ValueError("dt is null.")
    if venue_name is None:
        logging.warning("venue_name is null.")

    return (dt, teams, venue_name)


def _find_new_dt(
    soup: BeautifulSoup,
    scorebox_meta_div: Tag,
    url: str,
    session: ProxySession,
    league: League,
    player_urls: set[str],
    scores: list[float],
    fg: dict[str, int],
    fga: dict[str, int],
    offensive_rebounds: dict[str, int],
    assists: dict[str, int],
    turnovers: dict[str, int],
    positions_validator: dict[str, str],
) -> tuple[datetime.datetime, list[TeamModel], str]:
    in_divs = scorebox_meta_div.find_all("div")
    current_in_div_idx = 0
    in_div = in_divs[current_in_div_idx]
    in_div_text = in_div.get_text().strip()
    current_in_div_idx += 1
    if "Tournament" in in_div_text:
        in_div_text = in_divs[1].get_text().strip()
        current_in_div_idx += 1
    try:
        dt = parse(in_div_text)
    except dateutil.parser._parser.ParserError as exc:  # type: ignore
        logging.error("Failed to parse date for URL: %s", url)
        raise exc
    venue_div = in_divs[current_in_div_idx]
    venue_name = venue_div.get_text().strip()
    scorebox_div = soup.find("div", class_="scorebox")
    if not isinstance(scorebox_div, Tag):
        raise ValueError("scorebox_div is not a Tag.")

    teams: list[TeamModel] = []
    for a in scorebox_div.find_all("a"):
        team_url = urllib.parse.urljoin(url, a.get("href"))
        if "/schools/" in team_url:
            teams.append(
                create_sportsreference_team_model(
                    session,
                    team_url,
                    dt,
                    league,
                    player_urls,
                    scores[len(teams)],
                    fg,
                    fga,
                    offensive_rebounds,
                    assists,
                    turnovers,
                    a.get_text().strip(),
                    positions_validator,
                )
            )

    return (dt, teams, venue_name)


def _create_sportsreference_game_model(
    session: ProxySession,
    url: str,
    league: League,
    positions_validator: dict[str, str],
) -> GameModel:
    # pylint: disable=too-many-branches
    response = session.get(url)
    response.raise_for_status()
    soup = BeautifulSoup(response.text, "lxml")
    page_title = soup.find("h1", class_="page_title")

    # If the page_title is bad, try fetching from a non wayback source
    if page_title is not None:
        if "file not found" in page_title.get_text().strip().lower():
            session.cache.delete(urls=[url, response.url])
            with session.wayback_disabled():
                response = session.get(url)
            response.raise_for_status()
            soup = BeautifulSoup(response.text, "lxml")

    player_urls = set()
    for a in soup.find_all("a"):
        player_url = urllib.parse.urljoin(url, a.get("href"))
        if "/players/" in player_url and not player_url.endswith("/players/"):
            player_urls.add(player_url)

    scores = []
    for score_div in soup.find_all("div", class_="score"):
        try:
            scores.append(float(score_div.get_text().strip()))
        except ValueError as exc:
            logging.error(response.text)
            raise exc

    handle = io.StringIO()
    handle.write(response.text)
    handle.seek(0)
    fg = {}
    fga = {}
    offensive_rebounds = {}
    assists = {}
    turnovers = {}
    try:
        dfs = pd.read_html(handle)
        for df in dfs:
            if df.index.nlevels > 1:
                df.columns = df.columns.get_level_values(1)
            if "Starters" in df.columns.values:
                players = df["Starters"].tolist()
                if "FG" in df.columns.values:
                    fgs = df["FG"].tolist()
                    for idx, player in enumerate(players):
                        fg[player] = fgs[idx]
                if "FGA" in df.columns.values:
                    fgas = df["FGA"].tolist()
                    for idx, player in enumerate(players):
                        fga[player] = fgas[idx]
                if "OREB" in df.columns.values:
                    orebs = df["OREB"].tolist()
                    for idx, player in enumerate(players):
                        offensive_rebounds[player] = orebs[idx]
                if "AST" in df.columns.values:
                    asts = df["AST"].tolist()
                    for idx, player in enumerate(players):
                        assists[player] = asts[idx]
                if "TOV" in df.columns.values:
                    tovs = df["TOV"].tolist()
                    for idx, player in enumerate(players):
                        turnovers[player] = tovs[idx]
    except ValueError as exc:
        logging.error(url)
        logging.error(response.text)
        raise exc

    scorebox_meta_div = soup.find("div", class_="scorebox_meta")
    if not isinstance(scorebox_meta_div, Tag):
        dt, teams, venue_name = _find_old_dt(
            dfs,
            session,
            soup,
            url,
            league,
            player_urls,
            fg,
            fga,
            offensive_rebounds,
            assists,
            turnovers,
            response,
            positions_validator,
        )
    else:
        dt, teams, venue_name = _find_new_dt(
            soup,
            scorebox_meta_div,
            url,
            session,
            league,
            player_urls,
            scores,
            fg,
            fga,
            offensive_rebounds,
            assists,
            turnovers,
            positions_validator,
        )
    for team in teams:
        if team.name == "File Not Found":
            raise ValueError("team name is File Not Found (invalid)")

    season_type = SeasonType.REGULAR
    for h2 in soup.find_all("h2"):
        a = h2.find("a")
        if a is None:
            continue
        season_text = a.get_text().strip()
        match season_text:
            case "Big Sky Conference":
                season_type = SeasonType.REGULAR
            case "Big East Conference":
                season_type = SeasonType.REGULAR
            case "Big West Conference":
                season_type = SeasonType.REGULAR
            case "Big Ten Conference":
                season_type = SeasonType.REGULAR
            case "Mid-American Conference":
                season_type = SeasonType.REGULAR
            case "Horizon League":
                season_type = SeasonType.REGULAR
            case "Atlantic 10 Conference":
                season_type = SeasonType.REGULAR
            case "Mountain West Conference":
                season_type = SeasonType.REGULAR
            case "West Coast Conference":
                season_type = SeasonType.REGULAR
            case "American Athletic Conference":
                season_type = SeasonType.REGULAR
            case "Coastal Athletic Association":
                season_type = SeasonType.REGULAR
            case "Conference USA":
                season_type = SeasonType.REGULAR
            case "America East Conference":
                season_type = SeasonType.REGULAR
            case "Sun Belt Conference":
                season_type = SeasonType.REGULAR
            case "Metro Atlantic Athletic Conference":
                season_type = SeasonType.REGULAR
            case "Atlantic Sun Conference":
                season_type = SeasonType.REGULAR
            case "Ohio Valley Conference":
                season_type = SeasonType.REGULAR
            case "Mid-Eastern Athletic Conference":
                season_type = SeasonType.REGULAR
            case "Big South Conference":
                season_type = SeasonType.REGULAR
            case "Summit League":
                season_type = SeasonType.REGULAR
            case "Western Athletic Conference":
                season_type = SeasonType.REGULAR
            case "Big 12 Conference":
                season_type = SeasonType.REGULAR
            case "Southeastern Conference":
                season_type = SeasonType.REGULAR
            case "Patriot League":
                season_type = SeasonType.REGULAR
            case "Southern Conference":
                season_type = SeasonType.REGULAR
            case "Missouri Valley Conference":
                season_type = SeasonType.REGULAR
            case "Atlantic Coast Conference":
                season_type = SeasonType.REGULAR
            case "Southwest Athletic Conference":
                season_type = SeasonType.REGULAR
            case "Southland Conference":
                season_type = SeasonType.REGULAR
            case "Northeast Conference":
                season_type = SeasonType.REGULAR
            case "Ivy League":
                season_type = SeasonType.REGULAR
            case "NCAA Men's Tournament":
                season_type = SeasonType.REGULAR
            case "Pac-12 Conference":
                season_type = SeasonType.REGULAR
            case "Colonial Athletic Association":
                season_type = SeasonType.REGULAR
            case "Pacific-12 Conference":
                season_type = SeasonType.REGULAR
            case "NCAA Women's Tournament":
                season_type = SeasonType.REGULAR
            case "Pacific-10 Conference":
                season_type = SeasonType.REGULAR
            case "Great West Conference":
                season_type = SeasonType.REGULAR
            case _:
                logging.warning("Unrecognised Season Text: %s", season_text)
        break

    return GameModel(
        dt=dt,
        week=None,
        game_number=None,
        venue=create_sportsreference_venue_model(venue_name, session, dt),  # pyright: ignore
        teams=teams,
        league=str(league),
        year=dt.year,
        season_type=season_type,
        end_dt=None,
        attendance=None,
        postponed=None,
        play_off=None,
        distance=None,
        dividends=[],
        pot=None,
    )


@MEMORY.cache(ignore=["session"])
def _cached_create_sportsreference_game_model(
    session: ProxySession,
    url: str,
    league: League,
    positions_validator: dict[str, str],
) -> GameModel:
    return _create_sportsreference_game_model(session, url, league, positions_validator)


def create_sportsreference_game_model(
    session: ProxySession,
    url: str,
    league: League,
    positions_validator: dict[str, str],
) -> GameModel:
    """Create a sports reference game model."""
    if not pytest_is_running.is_running():
        return _cached_create_sportsreference_game_model(
            session, url, league, positions_validator
        )
    with session.cache_disabled():
        return _create_sportsreference_game_model(
            session, url, league, positions_validator
        )



"""The prototype class for odds."""

import datetime
from typing import Literal

from pydantic import BaseModel, Field

from .bookie_model import BookieModel
from .field_type import TYPE_KEY, FieldType

DT_COLUMN: Literal["dt"] = "dt"
ODDS_ODDS_COLUMN: Literal["odds"] = "odds"
ODDS_BOOKIE_COLUMN: Literal["bookie"] = "bookie"
ODDS_CANONICAL_COLUMN: Literal["canonical"] = "canonical"
ODDS_BET_COLUMN: Literal["bet"] = "bet"


class OddsModel(BaseModel):
    """The serialisable odds class."""

    odds: float = Field(
        ..., json_schema_extra={TYPE_KEY: FieldType.ODDS}, alias=ODDS_ODDS_COLUMN
    )
    bookie: BookieModel = Field(..., alias=ODDS_BOOKIE_COLUMN)
    dt: datetime.datetime | None = Field(
        ...,
        alias=DT_COLUMN,
    )
    canonical: bool = Field(..., alias=ODDS_CANONICAL_COLUMN)
    bet: str = Field(..., alias=ODDS_BET_COLUMN)



"""Tests for the sportsreference team model class."""
import datetime
import os
import unittest

import requests_mock
from sportsball.data.sportsreference.sportsreference_team_model import create_sportsreference_team_model
from sportsball.data.league import League
from sportsball.proxy_session import ProxySession


class TestSportsReferenceTeamModel(unittest.TestCase):

    def setUp(self):
        self.session = ProxySession(backend="memory")
        self.session._wayback_disabled = True
        self.dir = os.path.dirname(__file__)

    def test_identifier(self):
        with requests_mock.Mocker() as m:
            with open(os.path.join(self.dir, "2025.html"), "rb") as f:
                m.get("https://www.sports-reference.com/cbb/schools/villanova/men/2025.html", content=f.read())
            with open(os.path.join(self.dir, "kyle-neptune-1.html"), "rb") as f:
                m.get("https://www.sports-reference.com/cbb/coaches/kyle-neptune-1.html", content=f.read())
            m.get("https://news.google.com/rss/search?q=%22Villanova+Wildcats+Men%27s%22+%2B+%28sport+OR+ncaab+OR+%22NCAA+Division+I+Basketball%22%29+after%3A2010-10-08+before%3A2010-10-09&ceid=US:en&hl=en&gl=US")
            team_model = create_sportsreference_team_model(
                self.session,
                "https://www.sports-reference.com/cbb/schools/villanova/men/2025.html",
                datetime.datetime(2010, 10, 10, 10, 10, 00),
                League.NCAAB,
                set(),
                10.0,
                {},
                {},
                {},
                {},
                {},
                "",
                {},
            )
            self.assertEqual(team_model.identifier, "Villanova Wildcats Men's")

    def test_field_goals(self):
        url = "https://www.sports-reference.com/cbb/schools/villanova/men/2025.html"
        player_url = "https://www.sports-reference.com/cbb/players/eric-dixon-1.html"
        with requests_mock.Mocker() as m:
            with open(os.path.join(self.dir, "2025.html"), "rb") as f:
                m.get(url, content=f.read())
            with open(os.path.join(self.dir, "eric-dixon-1.html"), "rb") as f:
                m.get(player_url, content=f.read())
            with open(os.path.join(self.dir, "kyle-neptune-1.html"), "rb") as f:
                m.get("https://www.sports-reference.com/cbb/coaches/kyle-neptune-1.html", content=f.read())
            team_model = create_sportsreference_team_model(
                self.session,
                url,
                datetime.datetime(2010, 10, 10, 10, 10, 00),
                League.NCAAB,
                set([player_url]),
                10.0,
                {
                    "Eric Dixon": 8,
                },
                {},
                {},
                {},
                {},
                "",
                {},
            )
            self.assertEqual(team_model.field_goals, 8)

    def test_field_goals_attempted(self):
        url = "https://www.sports-reference.com/cbb/schools/villanova/men/2025.html"
        player_url = "https://www.sports-reference.com/cbb/players/eric-dixon-1.html"
        with requests_mock.Mocker() as m:
            with open(os.path.join(self.dir, "2025.html"), "rb") as f:
                m.get(url, content=f.read())
            with open(os.path.join(self.dir, "eric-dixon-1.html"), "rb") as f:
                m.get(player_url, content=f.read())
            with open(os.path.join(self.dir, "kyle-neptune-1.html"), "rb") as f:
                m.get("https://www.sports-reference.com/cbb/coaches/kyle-neptune-1.html", content=f.read())
            team_model = create_sportsreference_team_model(
                self.session,
                url,
                datetime.datetime(2010, 10, 10, 10, 10, 00),
                League.NCAAB,
                set([player_url]),
                10.0,
                {
                    "Eric Dixon": 8,
                },
                {
                    "Eric Dixon": 10
                },
                {},
                {},
                {},
                "",
                {},
            )
            self.assertEqual(team_model.field_goals_attempted, 10)

    def test_identifier_2(self):
        url = "http://www.basketball-reference.com/teams/LAC/2015.html"
        with requests_mock.Mocker() as m:
            with open(os.path.join(self.dir, "LAC_2015.html"), "rb") as f:
                m.get(url, content=f.read())
            with open(os.path.join(self.dir, "riverdo01c.html"), "rb") as f:
                m.get("http://www.basketball-reference.com/coaches/riverdo01c.html", content=f.read())
            team_model = create_sportsreference_team_model(
                self.session,
                url,
                datetime.datetime(2010, 10, 10, 10, 10, 00),
                League.NCAAB,
                set(),
                10.0,
                {},
                {},
                {},
                {},
                {},
                "",
                {},
            )
            self.assertEqual(team_model.identifier, "Los Angeles Clippers")

    def test_identifier_3(self):
        url = "http://www.basketball-reference.com/teams/BRK/2014.html"
        with requests_mock.Mocker() as m:
            with open(os.path.join(self.dir, "BRK_2014.html"), "rb") as f:
                m.get(url, content=f.read())
            with open(os.path.join(self.dir, "kiddja01c.html"), "rb") as f:
                m.get("http://www.basketball-reference.com/coaches/kiddja01c.html", content=f.read())
            team_model = create_sportsreference_team_model(
                self.session,
                url,
                datetime.datetime(2010, 10, 10, 10, 10, 00),
                League.NCAAB,
                set(),
                10.0,
                {},
                {},
                {},
                {},
                {},
                "",
                {},
            )
            self.assertEqual(team_model.identifier, "Brooklyn Nets")

    def test_name(self):
        url = "http://www.basketball-reference.com/teams/CLE/2016.html"
        with requests_mock.Mocker() as m:
            with open(os.path.join(self.dir, "CLE_2016.html"), "rb") as f:
                m.get(url, content=f.read())
            with open(os.path.join(self.dir, "luety01c.html"), "rb") as f:
                m.get("http://www.basketball-reference.com/coaches/luety01c.html", content=f.read())
            team_model = create_sportsreference_team_model(
                self.session,
                url,
                datetime.datetime(2010, 10, 10, 10, 10, 00),
                League.NCAAB,
                set(),
                10.0,
                {},
                {},
                {},
                {},
                {},
                "",
                {},
            )
            self.assertEqual(team_model.name, "Cleveland Cavaliers")

    def test_name_2(self):
        url = "http://www.basketball-reference.com/teams/LAL/2015.html"
        with requests_mock.Mocker() as m:
            with open(os.path.join(self.dir, "LAL_2015.html"), "rb") as f:
                m.get(url, content=f.read())
            with open(os.path.join(self.dir, "scottby01c.html"), "rb") as f:
                m.get("http://www.basketball-reference.com/coaches/scottby01c.html", content=f.read())
            team_model = create_sportsreference_team_model(
                self.session,
                url,
                datetime.datetime(2010, 10, 10, 10, 10, 00),
                League.NBA,
                set(),
                10.0,
                {},
                {},
                {},
                {},
                {},
                "",
                {},
            )
            self.assertEqual(team_model.name, "Los Angeles Lakers")

    def test_name_3(self):
        url = "https://www.basketball-reference.com/teams/BOS/2015.html"
        with requests_mock.Mocker() as m:
            with open(os.path.join(self.dir, "BOS_2015.html"), "rb") as f:
                m.get(url, content=f.read())
            with open(os.path.join(self.dir, "stevebr99c.html"), "rb") as f:
                m.get("https://www.basketball-reference.com/coaches/stevebr99c.html", content=f.read())
            team_model = create_sportsreference_team_model(
                self.session,
                url,
                datetime.datetime(2010, 10, 10, 10, 10, 00),
                League.NBA,
                set(),
                10.0,
                {},
                {},
                {},
                {},
                {},
                "",
                {},
            )
            self.assertEqual(team_model.name, "Boston Celtics")

    def test_coach_url_null(self):
        url = "https://www.basketball-reference.com/teams/GSW/2016.html"
        with requests_mock.Mocker() as m:
            with open(os.path.join(self.dir, "GSW_2016_2.html"), "rb") as f:
                m.get(url, content=f.read())
            with open(os.path.join(self.dir, "kerrst01c.html"), "rb") as f:
                m.get("https://www.basketball-reference.com/coaches/kerrst01c.html", content=f.read())
            team_model = create_sportsreference_team_model(
                session=self.session,
                url=url,
                dt=datetime.datetime(2010, 10, 10, 10, 10, 00),
                league=League.NBA,
                player_urls=set(),
                points=10.0,
                fg={},
                fga={},
                offensive_rebounds={},
                assists={},
                turnovers={},
                team_name="",
                positions_validator={},
            )
            self.assertListEqual(team_model.coaches, [])



"""Tests for the player model class."""
import unittest

from sportsball.data.player_model import PlayerModel
from sportsball.data.species import Species


class TestPlayerModel(unittest.TestCase):

    def setUp(self):
        self._player_model = PlayerModel(
            identifier="a",
            jersey="45",
            kicks=None,
            fumbles=None,
            fumbles_lost=None,
            field_goals=None,
            field_goals_attempted=10,
            offensive_rebounds=None,
            assists=None,
            turnovers=None,
            name="James Jones",
            marks=None,
            handballs=None,
            disposals=None,
            goals=None,
            behinds=None,
            hit_outs=None,
            tackles=None,
            rebounds=None,
            insides=None,
            clearances=None,
            clangers=None,
            free_kicks_for=None,
            free_kicks_against=None,
            brownlow_votes=None,
            contested_possessions=None,
            uncontested_possessions=None,
            contested_marks=None,
            marks_inside=None,
            one_percenters=None,
            bounces=None,
            goal_assists=None,
            percentage_played=None,
            birth_date=None,
            age=None,
            species=str(Species.HUMAN),
            handicap_weight=None,
            father=None,
            starting_position=None,
            weight=None,
            birth_address=None,
            owner=None,
        )

    def test_field_goals_attempted(self):
        self.assertEqual(self._player_model.field_goals_attempted, 10)



"""Tests for the afltables league model class."""
import unittest
import os
import datetime

import requests_cache
from sportsball.data.afl.afltables.afl_afltables_league_model import AFLAFLTablesLeagueModel, _find_dt
from sportsball.data.league import League


class TestAFLTablesLeagueModel(unittest.TestCase):

    def setUp(self):
        self._session = requests_cache.CachedSession(backend="memory")
        self._league_model = AFLAFLTablesLeagueModel(self._session)
        self.dir = os.path.dirname(__file__)

    def test_league(self):
        self.assertEqual(self._league_model.league, League.AFL)

    def test_new_find_dt(self):
        thing = "Richmond  Thu 13-Mar-2025 Venue: M.C.G."
        dt = _find_dt(thing, "https://afltables.com/afl/seas/2025.html")
        self.assertEqual(dt, datetime.datetime(2025, 3, 13, 0, 0))



"""An enumeration of the different season types."""

from enum import StrEnum


class SeasonType(StrEnum):
    """An enumeration for the season type."""

    REGULAR = "REGULAR"
    PRESEASON = "PRESEASON"
    POSTSEASON = "POSTSEASON"
    OFFSEASON = "OFFSEASON"



"""OddsPortal game model."""

# pylint: disable=too-many-locals,too-many-statements,line-too-long,broad-exception-caught,duplicate-code
import datetime
import json
import logging
import urllib.parse

import pytest_is_running
from bs4 import BeautifulSoup, Tag

from ...cache import MEMORY
from ...proxy_session import ProxySession
from ..game_model import GameModel
from ..league import League
from .decrypt import fetch_data
from .oddsportal_team_model import create_oddsportal_team_model
from .oddsportal_venue_model import create_oddsportal_venue_model


def _create_oddsportal_game_model(
    session: ProxySession,
    url: str,
    league: League,
) -> GameModel | None:
    with session.wayback_disabled():
        response = session.get(
            url,
        )
    response.raise_for_status()

    soup = BeautifulSoup(response.text, "lxml")
    event_tag = soup.find("div", id="react-event-header")
    if not isinstance(event_tag, Tag):
        soup_x = BeautifulSoup(response.text, "xml")
        event_tag = soup_x.find("Event")
        if not isinstance(event_tag, Tag):
            logging.error("event_tag is not a tag for URL: %s.", url)
            return None
        event = json.loads(str(event_tag[":data"]))
    else:
        event = json.loads(str(event_tag["data"]))

    event_body = event["eventBody"]
    event_data = event["eventData"]
    dt = datetime.datetime.fromtimestamp(event_body["startDate"])
    end_dt = None
    if event_body.get("endDate", False) is not False:
        end_dt = datetime.datetime.fromtimestamp(event_body["endDate"])
    version_id = str(event_data["versionId"])
    sport_id = str(event_data["sportId"])
    unique_id = event_data["id"]
    default_bet_id = str(event_data["defaultBetId"])
    default_scope_id = str(event_data["defaultScopeId"])
    xhash = urllib.parse.unquote(event_data["xhash"])

    bookie_names = event_body["providersNames"]
    home_points = None
    if event_body["homeResult"] != "":
        home_points = float(event_body["homeResult"])
    away_points = None
    if event_body["awayResult"] != "":
        away_points = float(event_body["awayResult"])

    venue = None
    if (
        event_body.get("venue", "")
        and event_body.get("venueTown", "")
        and event_body.get("venueCountry", "")
    ):
        venue = create_oddsportal_venue_model(
            session,
            dt,
            event_body["venue"],
            event_body["venueTown"],
            event_body["venueCountry"],
        )

    dat_url = f"https://www.oddsportal.com/match-event/{version_id}-{sport_id}-{unique_id}-{default_bet_id}-{default_scope_id}-{xhash}.dat"
    try:
        parsed_data = fetch_data(dat_url, session, url, soup)
    except Exception:
        return None

    return GameModel(
        dt=dt,
        week=None,
        game_number=None,
        venue=venue,
        teams=[
            create_oddsportal_team_model(
                session,
                dt,
                event_data["home"],
                league,
                home_points,
                default_bet_id,
                default_scope_id,
                bookie_names,
                0,
                parsed_data,
            ),
            create_oddsportal_team_model(
                session,
                dt,
                event_data["away"],
                league,
                away_points,
                default_bet_id,
                default_scope_id,
                bookie_names,
                1,
                parsed_data,
            ),
        ],
        end_dt=end_dt,
        attendance=None,
        league=str(league),
        year=dt.year,
        season_type=None,
        postponed=event_data["isPostponed"],
        play_off=None,
        distance=None,
        dividends=[],
        pot=None,
    )


@MEMORY.cache(ignore=["session"])
def _cached_create_oddsportal_game_model(
    session: ProxySession,
    url: str,
    league: League,
) -> GameModel | None:
    return _create_oddsportal_game_model(session, url, league)


def create_oddsportal_game_model(
    session: ProxySession,
    url: str,
    league: League,
    is_next: bool,
) -> GameModel | None:
    """Create a OddsPortal game model."""
    if not pytest_is_running.is_running() and not is_next:
        return _cached_create_oddsportal_game_model(session, url, league)
    with session.cache_disabled():
        return _create_oddsportal_game_model(session, url, league)



"""Tests for the sportsreference league model class."""
import unittest
import os

import requests_cache
from sportsball.data.sportsreference.sportsreference_league_model import SportsReferenceLeagueModel, _find_game_urls
from sportsball.data.league import League
from bs4 import BeautifulSoup


class TestSportsReferenceLeagueModel(unittest.TestCase):

    def setUp(self):
        self.session = requests_cache.CachedSession(backend="memory")
        self.league_model = SportsReferenceLeagueModel(self.session, League.NBA, "https://www.basketball-reference.com/boxscores/")
        self.dir = os.path.dirname(__file__)

    def test_league(self):
        self.assertEqual(self.league_model.league, League.NBA)

    def test_day(self):
        url = "https://www.sports-reference.com/cbb/boxscores/index.cgi?month=11&day=22&year=2022"
        with open(os.path.join(self.dir, "22_11_2022.html")) as handle:
            soup = BeautifulSoup(handle.read(), "lxml")
            game_urls = _find_game_urls(soup, url)
            self.assertListEqual(game_urls, [
                'https://www.sports-reference.com/cbb/boxscores/2022-11-22-20-cal-poly_w.html',
                'https://www.sports-reference.com/cbb/boxscores/2022-11-22-20-arkansas.html',
                'https://www.sports-reference.com/cbb/boxscores/2022-11-22-18-auburn.html',
                'https://www.sports-reference.com/cbb/boxscores/2022-11-22-22-arizona.html',
                'https://www.sports-reference.com/cbb/boxscores/2022-11-22-20-nebraska-omaha_w.html',
                'https://www.sports-reference.com/cbb/boxscores/2022-11-22-14-louisville.html',
                'https://www.sports-reference.com/cbb/boxscores/2022-11-22-00-yale.html',
                'https://www.sports-reference.com/cbb/boxscores/2022-11-22-01-arkansas-pine-bluff.html',
                'https://www.sports-reference.com/cbb/boxscores/2022-11-22-01-hampton.html',
                'https://www.sports-reference.com/cbb/boxscores/2022-11-22-01-jacksonville-state.html',
                'https://www.sports-reference.com/cbb/boxscores/2022-11-22-01-southern-illinois.html',
                'https://www.sports-reference.com/cbb/boxscores/2022-11-22-02-arizona-state.html',
                'https://www.sports-reference.com/cbb/boxscores/2022-11-22-02-utah-state.html',
                'https://www.sports-reference.com/cbb/boxscores/2022-11-22-03-sacramento-state.html',
                'https://www.sports-reference.com/cbb/boxscores/2022-11-22-11-austin-peay.html',
                'https://www.sports-reference.com/cbb/boxscores/2022-11-22-11-austin-peay_w.html',
                'https://www.sports-reference.com/cbb/boxscores/2022-11-22-11-bucknell_w.html',
                'https://www.sports-reference.com/cbb/boxscores/2022-11-22-11-east-carolina.html',
                'https://www.sports-reference.com/cbb/boxscores/2022-11-22-11-grand-canyon.html',
                'https://www.sports-reference.com/cbb/boxscores/2022-11-22-11-illinois-state.html',
                'https://www.sports-reference.com/cbb/boxscores/2022-11-22-12-eastern-michigan.html',
                'https://www.sports-reference.com/cbb/boxscores/2022-11-22-12-northern-illinois.html',
                'https://www.sports-reference.com/cbb/boxscores/2022-11-22-13-akron.html',
                'https://www.sports-reference.com/cbb/boxscores/2022-11-22-13-northern-kentucky.html',
                'https://www.sports-reference.com/cbb/boxscores/2022-11-22-13-sacred-heart.html',
                'https://www.sports-reference.com/cbb/boxscores/2022-11-22-13-wyoming_w.html',
                'https://www.sports-reference.com/cbb/boxscores/2022-11-22-14-colgate_w.html',
                'https://www.sports-reference.com/cbb/boxscores/2022-11-22-14-davidson_w.html',
                'https://www.sports-reference.com/cbb/boxscores/2022-11-22-14-elon_w.html',
                'https://www.sports-reference.com/cbb/boxscores/2022-11-22-14-long-island-university.html',
                'https://www.sports-reference.com/cbb/boxscores/2022-11-22-14-mississippi-valley-state_w.html',
                'https://www.sports-reference.com/cbb/boxscores/2022-11-22-14-san-francisco.html',
                'https://www.sports-reference.com/cbb/boxscores/2022-11-22-15-winthrop.html',
                'https://www.sports-reference.com/cbb/boxscores/2022-11-22-16-detroit-mercy_w.html',
                'https://www.sports-reference.com/cbb/boxscores/2022-11-22-16-georgia.html',
                'https://www.sports-reference.com/cbb/boxscores/2022-11-22-16-holy-cross.html',
                'https://www.sports-reference.com/cbb/boxscores/2022-11-22-16-marshall_w.html',
                'https://www.sports-reference.com/cbb/boxscores/2022-11-22-17-cincinnati.html',
                'https://www.sports-reference.com/cbb/boxscores/2022-11-22-17-cornell.html',
                'https://www.sports-reference.com/cbb/boxscores/2022-11-22-17-fresno-state_w.html',
                'https://www.sports-reference.com/cbb/boxscores/2022-11-22-17-indiana-state.html',
                'https://www.sports-reference.com/cbb/boxscores/2022-11-22-17-jacksonville-state_w.html',
                'https://www.sports-reference.com/cbb/boxscores/2022-11-22-17-maryland-eastern-shore_w.html',
                'https://www.sports-reference.com/cbb/boxscores/2022-11-22-17-morehead-state_w.html',
                'https://www.sports-reference.com/cbb/boxscores/2022-11-22-17-morgan-state_w.html',
                'https://www.sports-reference.com/cbb/boxscores/2022-11-22-17-north-carolina-central_w.html',
                'https://www.sports-reference.com/cbb/boxscores/2022-11-22-17-queens-nc_w.html',
                'https://www.sports-reference.com/cbb/boxscores/2022-11-22-17-rhode-island.html',
                'https://www.sports-reference.com/cbb/boxscores/2022-11-22-17-south-carolina-state_w.html',
                'https://www.sports-reference.com/cbb/boxscores/2022-11-22-17-southern-methodist_w.html',
                'https://www.sports-reference.com/cbb/boxscores/2022-11-22-18-arkansas-state.html',
                'https://www.sports-reference.com/cbb/boxscores/2022-11-22-18-bethune-cookman_w.html',
                'https://www.sports-reference.com/cbb/boxscores/2022-11-22-18-buffalo_w.html',
                'https://www.sports-reference.com/cbb/boxscores/2022-11-22-18-central-florida_w.html',
                'https://www.sports-reference.com/cbb/boxscores/2022-11-22-18-drexel_w.html',
                'https://www.sports-reference.com/cbb/boxscores/2022-11-22-18-duquesne_w.html',
                'https://www.sports-reference.com/cbb/boxscores/2022-11-22-18-hartford.html',
                'https://www.sports-reference.com/cbb/boxscores/2022-11-22-18-lafayette_w.html',
                'https://www.sports-reference.com/cbb/boxscores/2022-11-22-18-massachusetts-lowell_w.html',
                'https://www.sports-reference.com/cbb/boxscores/2022-11-22-18-north-carolina-at.html',
                'https://www.sports-reference.com/cbb/boxscores/2022-11-22-18-notre-dame.html',
                'https://www.sports-reference.com/cbb/boxscores/2022-11-22-18-rhode-island_w.html',
                'https://www.sports-reference.com/cbb/boxscores/2022-11-22-18-saint-josephs.html',
                'https://www.sports-reference.com/cbb/boxscores/2022-11-22-18-texas-am-corpus-christi.html',
                'https://www.sports-reference.com/cbb/boxscores/2022-11-22-19-abilene-christian.html',
                'https://www.sports-reference.com/cbb/boxscores/2022-11-22-19-alcorn-state_w.html',
                'https://www.sports-reference.com/cbb/boxscores/2022-11-22-19-drexel.html',
                'https://www.sports-reference.com/cbb/boxscores/2022-11-22-19-fordham.html',
                'https://www.sports-reference.com/cbb/boxscores/2022-11-22-19-george-washington.html',
                'https://www.sports-reference.com/cbb/boxscores/2022-11-22-19-incarnate-word.html',
                'https://www.sports-reference.com/cbb/boxscores/2022-11-22-19-kansas-state.html',
                'https://www.sports-reference.com/cbb/boxscores/2022-11-22-19-long-island-university_w.html',
                'https://www.sports-reference.com/cbb/boxscores/2022-11-22-19-loyola-il_w.html',
                'https://www.sports-reference.com/cbb/boxscores/2022-11-22-19-morehead-state.html',
                'https://www.sports-reference.com/cbb/boxscores/2022-11-22-19-navy_w.html',
                'https://www.sports-reference.com/cbb/boxscores/2022-11-22-19-nebraska_w.html',
                'https://www.sports-reference.com/cbb/boxscores/2022-11-22-19-norfolk-state.html',
                'https://www.sports-reference.com/cbb/boxscores/2022-11-22-19-norfolk-state_w.html',
                'https://www.sports-reference.com/cbb/boxscores/2022-11-22-19-north-alabama_w.html',
                'https://www.sports-reference.com/cbb/boxscores/2022-11-22-19-north-carolina-greensboro.html',
                'https://www.sports-reference.com/cbb/boxscores/2022-11-22-19-north-carolina-wilmington_w.html',
                'https://www.sports-reference.com/cbb/boxscores/2022-11-22-19-pacific_w.html',
                'https://www.sports-reference.com/cbb/boxscores/2022-11-22-19-radford_w.html',
                'https://www.sports-reference.com/cbb/boxscores/2022-11-22-19-richmond.html',
                'https://www.sports-reference.com/cbb/boxscores/2022-11-22-19-rutgers.html',
                'https://www.sports-reference.com/cbb/boxscores/2022-11-22-19-samford_w.html',
                'https://www.sports-reference.com/cbb/boxscores/2022-11-22-19-south-florida_w.html',
                'https://www.sports-reference.com/cbb/boxscores/2022-11-22-19-southeast-missouri-state.html',
                'https://www.sports-reference.com/cbb/boxscores/2022-11-22-19-southern-illinois_w.html',
                'https://www.sports-reference.com/cbb/boxscores/2022-11-22-19-st-bonaventure.html',
                'https://www.sports-reference.com/cbb/boxscores/2022-11-22-19-tennessee-tech.html',
                'https://www.sports-reference.com/cbb/boxscores/2022-11-22-19-texas-pan-american.html',
                'https://www.sports-reference.com/cbb/boxscores/2022-11-22-19-towson.html',
                'https://www.sports-reference.com/cbb/boxscores/2022-11-22-19-western-michigan_w.html',
                'https://www.sports-reference.com/cbb/boxscores/2022-11-22-19-xavier_w.html',
                'https://www.sports-reference.com/cbb/boxscores/2022-11-22-20-albany-ny.html',
                'https://www.sports-reference.com/cbb/boxscores/2022-11-22-20-arkansas-state_w.html',
                'https://www.sports-reference.com/cbb/boxscores/2022-11-22-20-bradley_w.html',
                'https://www.sports-reference.com/cbb/boxscores/2022-11-22-20-california-davis_w.html',
                'https://www.sports-reference.com/cbb/boxscores/2022-11-22-20-chicago-state_w.html',
                'https://www.sports-reference.com/cbb/boxscores/2022-11-22-20-lafayette.html',
                'https://www.sports-reference.com/cbb/boxscores/2022-11-22-20-maryland-eastern-shore.html',
                'https://www.sports-reference.com/cbb/boxscores/2022-11-22-20-nicholls-state.html',
                'https://www.sports-reference.com/cbb/boxscores/2022-11-22-20-north-dakota.html',
                'https://www.sports-reference.com/cbb/boxscores/2022-11-22-20-north-dakota-state_w.html',
                'https://www.sports-reference.com/cbb/boxscores/2022-11-22-20-north-texas.html',
                'https://www.sports-reference.com/cbb/boxscores/2022-11-22-20-northern-colorado.html',
                'https://www.sports-reference.com/cbb/boxscores/2022-11-22-20-northwestern.html',
                'https://www.sports-reference.com/cbb/boxscores/2022-11-22-20-northwestern_w.html',
                'https://www.sports-reference.com/cbb/boxscores/2022-11-22-20-pittsburgh.html',
                'https://www.sports-reference.com/cbb/boxscores/2022-11-22-20-south-alabama_w.html',
                'https://www.sports-reference.com/cbb/boxscores/2022-11-22-20-southern-methodist.html',
                'https://www.sports-reference.com/cbb/boxscores/2022-11-22-20-texas-san-antonio.html',
                'https://www.sports-reference.com/cbb/boxscores/2022-11-22-21-idaho-state_w.html',
                'https://www.sports-reference.com/cbb/boxscores/2022-11-22-21-montana.html',
                'https://www.sports-reference.com/cbb/boxscores/2022-11-22-21-new-mexico_w.html',
                'https://www.sports-reference.com/cbb/boxscores/2022-11-22-21-pepperdine_w.html',
                'https://www.sports-reference.com/cbb/boxscores/2022-11-22-21-san-diego-state_w.html',
                'https://www.sports-reference.com/cbb/boxscores/2022-11-22-21-san-jose-state_w.html',
                'https://www.sports-reference.com/cbb/boxscores/2022-11-22-21-santa-clara_w.html',
                'https://www.sports-reference.com/cbb/boxscores/2022-11-22-21-texas-el-paso.html',
                'https://www.sports-reference.com/cbb/boxscores/2022-11-22-22-cal-state-fullerton.html',
                'https://www.sports-reference.com/cbb/boxscores/2022-11-22-22-california-riverside.html',
                'https://www.sports-reference.com/cbb/boxscores/2022-11-22-22-pacific.html',
            ])



"""The prototype class for an owner."""

from typing import Literal

from pydantic import BaseModel, Field

from .field_type import TYPE_KEY, FieldType

OWNER_IDENTIFIER_COLUMN: Literal["identifier"] = "identifier"
OWNER_NAME_COLUMN: Literal["name"] = "name"


class OwnerModel(BaseModel):
    """The serialisable owner class."""

    identifier: str = Field(
        ...,
        json_schema_extra={TYPE_KEY: FieldType.CATEGORICAL},
        alias=OWNER_IDENTIFIER_COLUMN,
    )
    name: str = Field(..., alias=OWNER_NAME_COLUMN)



"""NCAAB combined league model."""

from ....proxy_session import ProxySession
from ...combined.combined_league_model import CombinedLeagueModel
from ...league import League
from ..espn.ncaab_espn_league_model import NCAABESPNLeagueModel
from ..oddsportal.ncaab_oddsportal_league_model import \
    NCAABOddsPortalLeagueModel
from ..sportsreference.ncaab_sportsreference_league_model import \
    NCAABSportsReferenceLeagueModel

DEPAUL_BLUE_DEMONS = "DePaul Blue Demons Men's"
GEORGETOWN_HOYAS = "Georgetown Hoyas Men's"
CAL_STATE_NORTHRIDGE_MATADORS = "Cal State Northridge Matadors Men's"
HAWAII_RAINBOW_WARRIORS = "Hawaii Rainbow Warriors Men's"
INDIANA_HOOSIERS = "Indiana Hoosiers Men's"
OHIO_STATE_BUCKEYES = "Ohio State Buckeyes Men's"
OHIO_BOBCATS = "Ohio Bobcats Men's"
AKRON_ZIPS = "Akron Zips Men's"
ROBERT_MORRIS_COLONIALS = "Robert Morris Colonials Men's"
GREEN_BAY_PHOENIX = "Green Bay Phoenix Men's"
VCU_RAMS = "VCU Rams Men's"
ST_JOSEPHS_HAWKS = "St. Joseph's Hawks Men's"
PROVIDENCE_FRIARS = "Providence Friars Men's"
VILLANOVA_WILDCATS = "Villanova Wildcats Men's"
YOUNGSTOWN_STATE_PENGUINS = "Youngstown State Penguins Men's"
MILWAUKEE_PANTHERS = "Milwaukee Panthers Men's"
IOWA_HAWKEYES = "Iowa Hawkeyes Men's"
UCLA_BRUINS = "UCLA Bruins Men's"
AIR_FORCE_FALCONS = "Air Force Falcons Men's"
FRESNO_STATE_BULLDOGS = "Fresno State Bulldogs Men's"
BOISE_STATE_BRONCOS = "Boise State Broncos Men's"
NEW_MEXICO_LOBOS = "New Mexico Lobos Men's"
GONZAGA_BULLDOGS = "Gonzaga Bulldogs Men's"
OREGON_STATE_BEAVERS = "Oregon State Beavers Men's"
MEMPHIS_TIGERS = "Memphis Tigers Men's"
TEMPLE_OWLS = "Temple Owls Men's"
MICHIGAN_WOLVERINES = "Michigan Wolverines Men's"
MINNESOTA_GOLDEN_GOPHERS = "Minnesota Golden Gophers Men's"
COLLEGE_CHARLESTON_COUGARS = "College of Charleston Cougars Men's"
CAMPBELL_FIGHTING_CAMELS = "Campbell Fighting Camels Men's"
STONY_BROOKS_SEAWOLVES = "Stony Brook Seawolves Men's"
DELAWARE_FIGHTIN_BLUE_HENS = "Delaware Fightin' Blue Hens Men's"
NEW_MEXICO_STATE_AGGIES = "New Mexico State Aggies Men's"
FLORIDA_INTERNATIONAL_PANTHERS = "Florida International Panthers Men's"
NEW_HAMPSHIRE_WILDCATS = "New Hampshire Wildcats Men's"
UMBC_RETRIEVERS = "UMBC Retrievers Men's"
APPALACHIAN_STATE_MOUNTAINEERS = "Appalachian State Mountaineers Men's"
OLD_DOMINION_MONARCHS = "Old Dominion Monarchs Men's"
MARIST_RED_FOXES = "Marist Red Foxes Men's"
ST_PETERS_PEACOCKS = "St. Peter's Peacocks Men's"
QUEENS_ROYALS = "Queens (NC) Royals Men's"
STETSON_HATTERS = "Stetson Hatters Men's"
EASTERN_KENTUCKY_COLONELS = "Eastern Kentucky Colonels Men's"
AUSTIN_PEAY_GOVERNORS = "Austin Peay Governors Men's"
RUTGERS_SCARLET_KNIGHTS = "Rutgers Scarlet Knights Men's"
NEBRASKA_CORNHUSKERS = "Nebraska Cornhuskers Men's"
MARYLAND_TERRAPINS = "Maryland Terrapins Men's"
NORTHWESTERN_WILDCATS = "Northwestern Wildcats Men's"
SIU_EDWARDSVILLE_COUGARS = "SIU-Edwardsville Cougars Men's"
TENNESSEE_STATE_TIGERS = "Tennessee State Tigers Men's"
SOUTHEAST_MISSOURI_STATE_REDHAWKS = "Southeast Missouri State Redhawks Men's"
TENNESSEE_TECH_GOLDEN_EAGLES = "Tennessee Tech Golden Eagles Men's"
UC_RIVERSIDE_HIGHLANDERS = "UC-Riverside Highlanders Men's"
UCSB_GAUCHOS = "UCSB Gauchos Men's"
WASHINGTON_STATE_COUGARS = "Washington State Cougars Men's"
SAN_DIEGO_TOREROS = "San Diego Toreros Men's"
MERRIMACK_WARRIORS = "Merrimack Warriors Men's"
QUINNIPIAC_BOBCATS = "Quinnipiac Bobcats Men's"
COPPIN_STATE_EAGLES = "Coppin State Eagles Men's"
MAINE_BLACK_BEARS = "Maine Black Bears Men's"
NJIT_HIGHLANDERS = "NJIT Highlanders Men's"
BRYANT_BULLDOGS = "Bryant Bulldogs Men's"
ALBANY_NY_GREAT_DANES = "Albany (NY) Great Danes Men's"
LIPSCOMB_BISONS = "Lipscomb Bisons Men's"
BELLARMINE_KNIGHTS = "Bellarmine Knights Men's"
WEST_GEORGIA_WOLVES = "West Georgia Wolves Men's"
FLORIDA_GULF_COAST_EAGLES = "Florida Gulf Coast Eagles Men's"
GEORGIA_SOUTHERN_EAGLES = "Georgia Southern Eagles Men's"
COASTAL_CAROLINA_CHANTICLEERS = "Coastal Carolina Chanticleers Men's"
ELON_PHOENIX = "Elon Phoenix Men's"
DREXEL_DRAGONS = "Drexel Dragons Men's"
NIAGARA_PURPLE_EAGLES = "Niagara Purple Eagles Men's"
FAIRFIELD_STAGS = "Fairfield Stags Men's"
WILLIAM_AND_MARY_TRIBE = "William & Mary Tribe Men's"
HAMPTON_PIRATES = "Hampton Pirates Men's"
CANISIUS_GOLDEN_GRIFFINS = "Canisius Golden Griffins Men's"
IONA_GAELS = "Iona Gaels Men's"
NORTH_ALABAMA_LIONS = "North Alabama Lions Men's"
JACKSONVILLE_DOLPHINS = "Jacksonville Dolphins Men's"
LOUISIANA_TECH_BULLDOGS = "Louisiana Tech Bulldogs Men's"
JACKSONVILLE_STATE_GAMECOCKS = "Jacksonville State Gamecocks Men's"
MARSHALL_THUNDERING_HERD = "Marshall Thundering Herd Men's"
JAMES_MADISON_DUKES = "James Madison Dukes Men's"
SAM_HOUSTON_BEARKATS = "Sam Houston Bearkats Men's"
KENNESAW_STATE_OWLS = "Kennesaw State Owls Men's"
HIGH_POINT_PANTHERS = "High Point Panthers Men's"
LONGWOOD_LANCERS = "Longwood Lancers Men's"
WESTERN_ILLINOIS_LEATHERNECKS = "Western Illinois Leathernecks Men's"
MOREHEAD_STATE_EAGLES = "Morehead State Eagles Men's"
MONMOUTH_HAWKS = "Monmouth Hawks Men's"
NORTH_CAROLINA_ATT_AGGIES = "North Carolina A&T Aggies Men's"
NORTHEASTERN_HUSKIES = "Northeastern Huskies Men's"
UNC_WILMINGTON_SEAHAWKS = "UNC Wilmington Seahawks Men's"
CENTRAL_ARKANSAS_BEARS = "Central Arkansas Bears Men's"
NORTH_FLORIDA_OSPREYS = "North Florida Ospreys Men's"
MOUNT_ST_MARYS_MOUNTAINEERS = "Mount St. Mary's Mountaineers Men's"
RIDER_BRONCS = "Rider Broncs Men's"
SACRED_HEART_PIONEERS = "Sacred Heart Pioneers Men's"
SIENA_SAINTS = "Siena Saints Men's"
HOFSTRA_PRIDE = "Hofstra Pride Men's"
TOWSON_TIGERS = "Towson Tigers Men's"
BINGHAMTON_BEARCATS = "Binghamton Bearcats Men's"
VERMONT_CATAMOUNTS = "Vermont Catamounts Men's"
LOUISIANA_RAGIN_CAJUNS = "Louisiana Ragin' Cajuns Men's"
ARKANSAS_STATE_RED_WOLVES = "Arkansas State Red Wolves Men's"
MONTANA_STATE_BOBCATS = "Montana State Bobcats Men's"
IDAHO_STATE_BENGALS = "Idaho State Bengals Men's"
KANSAS_CITY_ROOS = "Kansas City Roos Men's"
NORTH_DAKOTA_STATE_BISON = "North Dakota State Bison Men's"
SACRAMENTO_STATE_HORNETS = "Sacramento State Hornets Men's"
NORTHERN_ARIZONA_LUMBERJACKS = "Northern Arizona Lumberjacks Men's"
ORAL_ROBERTS_GOLDEN_EAGLES = "Oral Roberts Golden Eagles Men's"
SOUTH_DAKOTA_COYOTES = "South Dakota Coyotes Men's"
NORTH_DAKOTA_FIGHTING_HAWKS = "North Dakota Fighting Hawks Men's"
SOUTH_DAKOTA_STATE_JACKRABBITS = "South Dakota State Jackrabbits Men's"
LINDENWOOD_LIONS = "Lindenwood Lions Men's"
SOUTHERN_INDIANA_SCREAMING_EAGLES = "Southern Indiana Screaming Eagles Men's"
CALIFORNIA_BAPTIST_LANCERS = "California Baptist Lancers Men's"
TARLETON_STATE_TEXANS = "Tarleton State Texans Men's"
EASTERN_ILLINOIS_PANTHERS = "Eastern Illinois Panthers Men's"
UT_MARTIN_SKYHAWKS = "UT-Martin Skyhawks Men's"
UTAH_VALLEY_WOLVERINES = "Utah Valley Wolverines Men's"
UT_ARLINGTON_MAVERICKS = "UT Arlington Mavericks Men's"
ABILENE_CHRISTIAN_WILDCATS = "Abilene Christian Wildcats Men's"
GRAND_CANYON_ANTELOPES = "Grand Canyon Antelopes Men's"
UTEP_MINERS = "UTEP Miners Men's"
LIBERTY_FLAMES = "Liberty Flames Men's"
PORTLAND_STATE_VIKINGS = "Portland State Vikings Men's"
NORTHERN_COLORADO_BEARS = "Northern Colorado Bears Men's"
MONTANA_GRIZZLIES = "Montana Grizzlies Men's"
WEBER_STATE_WILDCATS = "Weber State Wildcats Men's"
UC_DAVIS_AGGIES = "UC-Davis Aggies Men's"
CAL_POLY_MUSTANGS = "Cal Poly Mustangs Men's"
CAL_STATE_FULLERTON_TITANS = "Cal State Fullerton Titans Men's"
UC_IRVINE_ANTEATERS = "UC-Irvine Anteaters Men's"
UC_SAN_DIEGO_TRITONS = "UC-San Diego Tritons Men's"
LONG_BEACH_STATE_BEACH = "Long Beach State Beach Men's"
SANTA_CLARA_BRONCOS = "Santa Clara Broncos Men's"
LOYOLA_MARYMOUNT_LIONS = "Loyola Marymount Lions Men's"
SAN_FRANCISCO_DONS = "San Francisco Dons Men's"
PEPPERDINE_WAVES = "Pepperdine Waves Men's"
PACIFIC_TIGERS = "Pacific Tigers Men's"
PORTLAND_PILOTS = "Portland Pilots Men's"
UTAH_TECH_TRAILBLAZERS = "Utah Tech Trailblazers Men's"
SEATTLE_REDHAWKS = "Seattle Redhawks Men's"
KANSAS_JAYHAWKS = "Kansas Jayhawks Men's"
IOWA_STATE_CYCLONES = "Iowa State Cyclones Men's"
GEORGIA_BULLDOGS = "Georgia Bulldogs Men's"
TENNESSEE_VOLUNTEERS = "Tennessee Volunteers Men's"
WEST_VIRGINIA_MOUNTAINEERS = "West Virginia Mountaineers Men's"
HOUSTON_COUGARS = "Houston Cougars Men's"
PENN_STATE_NITTANY_LIONS = "Penn State Nittany Lions Men's"
MICHIGAN_STATE_SPARTANS = "Michigan State Spartans Men's"
PURDUE_BOILERMAKERS = "Purdue Boilermakers Men's"
WASHINGTON_HUSKIES = "Washington Huskies Men's"
UTAH_STATE_AGGIES = "Utah State Aggies Men's"
UNLV_REBELS = "UNLV Rebels Men's"
LAFAYETTE_LEOPARDS = "Lafayette Leopards Men's"
ARMY_BLACK_KNIGHTS = "Army Black Knights Men's"
WOFFORD_TERRIERS = "Wofford Terriers Men's"
CHATTANOOGA_MOCS = "Chattanooga Mocs Men's"
GEORGE_MASON_PATRIOTS = "George Mason Patriots Men's"
DAYTON_FLYERS = "Dayton Flyers Men's"
DAVIDSON_WILDCATS = "Davidson Wildcats Men's"
LA_SALLE_EXPLORERS = "La Salle Explorers Men's"
SAMFORD_BULLDOGS = "Samford Bulldogs Men's"
MERCER_BEARS = "Mercer Bears Men's"
GARDNER_WEBB_RUNNIN_BULLDOGS = "Gardner-Webb Runnin' Bulldogs Men's"
RADFORD_HIGHLANDERS = "Radford Highlanders Men's"
ILLINOIS_STATE_REDBIRDS = "Illinois State Redbirds Men's"
DRAKE_BULLDOGS = "Drake Bulldogs Men's"
RHODE_ISLAND_RAMS = "Rhode Island Rams Men's"
LOYOLA_IL_RAMBLERS = "Loyola (IL) Ramblers Men's"
SOUTHERN_ILLINOIS_SALUKIS = "Southern Illinois Salukis Men's"
MISSOURI_STATE_BEARS = "Missouri State Bears Men's"
LEHIGH_MOUNTAIN_HAWKS = "Lehigh Mountain Hawks Men's"
BOSTON_UNIVERSITY_TERRIERS = "Boston University Terriers Men's"
USC_UPSTATE_SPARTANS = "USC Upstate Spartans Men's"
UNC_ASHEVILLE_BULLDOGS = "UNC Asheville Bulldogs Men's"
SOUTH_CAROLINA_GAMECOCKS = "South Carolina Gamecocks Men's"
VANDERBILT_COMMODORES = "Vanderbilt Commodores Men's"
WESTERN_CAROLINA_CATAMOUNTS = "Western Carolina Catamounts Men's"
VMI_KEYDETS = "VMI Keydets Men's"
STANFORD_CARDINAL = "Stanford Cardinal Men's"
WAKE_FOREST_DEMON_DEACONS = "Wake Forest Demon Deacons Men's"
CHARLESTON_SOUTHERN_BUCCANEERS = "Charleston Southern Buccaneers Men's"
WINTHROP_EAGLES = "Winthrop Eagles Men's"
HOLY_CROSS_CRUSADERS = "Holy Cross Crusaders Men's"
BUCKNELL_BISON = "Bucknell Bison Men's"
SETON_HALL_PIRATES = "Seton Hall Pirates Men's"
BUTLER_BULLDOGS = "Butler Bulldogs Men's"
NORTHERN_KENTUCKY_NORSE = "Northern Kentucky Norse Men's"
CLEVELAND_STATE_VIKINGS = "Cleveland State Vikings Men's"
UMASS_MINUTEMEN = "UMass Minutemen Men's"
FORDHAM_RAMS = "Fordham Rams Men's"
ETSU_BUCCANEERS = "ETSU Buccaneers Men's"
FURMANS_PALADINS = "Furman Paladins Men's"
DUQUESNE_DUKES = "Duquesne Dukes Men's"
GEORGE_WASHINGTON_REVOLUTIONARIES = "George Washington Revolutionaries Men's"
HOWARD_BISON = "Howard Bison Men's"
TROY_TROJANS = "Troy Trojans Men's"
LOUISIANA_MONRE_WARHAWKS = "Louisiana-Monroe Warhawks Men's"
COLGATE_RAIDERS = "Colgate Raiders Men's"
NAVY_MIDSHIPMEN = "Navy Midshipmen Men's"
CALIFORNIA_GOLDEN_BEARS = "California Golden Bears Men's"
UNC_TAR_HEELS = "UNC Tar Heels Men's"
THE_CITADEL_BULLDOGS = "The Citadel Bulldogs Men's"
UNC_GREENSBORO_SPARTANS = "UNC Greensboro Spartans Men's"
IU_INDIANAPOLIS_JAGUARS = "IU Indianapolis Jaguars Men's"
OAKLAND_GOLDEN_GRIZZLIES = "Oakland Golden Grizzlies Men's"
UAB_BLAZERS = "UAB Blazers Men's"
SOUTH_FLORIDA_BULLS = "South Florida Bulls Men's"
RICHMOND_SPIDERS = "Richmond Spiders Men's"
ST_BONAVENTURE_BONNIES = "St. Bonaventure Bonnies Men's"
FLORIDA_ATLANTIC_OWLS = "Florida Atlantic Owls Men's"
TULANE_GREEN_WAVE = "Tulane Green Wave Men's"
NC_STATE_WOLFPACK = "NC State Wolfpack Men's"
VIRGINIA_TECH_HOKIES = "Virginia Tech Hokies Men's"
PURDUE_FORT_WAYNE_MASTODONS = "Purdue Fort Wayne Mastodons Men's"
WRIGHT_STATE_RAIDERS = "Wright State Raiders Men's"
INDIANA_STATE_SYCAMORES = "Indiana State Sycamores Men's"
BRADLEY_BRAVES = "Bradley Braves Men's"
MURRAY_STATE_RACERS = "Murray State Racers Men's"
UIC_FLAMES = "UIC Flames Men's"
SOUTHERN_MISS_GOLDEN_EAGLES = "Southern Miss Golden Eagles Men's"
SOUTH_ALABAMA_JAGUARS = "South Alabama Jaguars Men's"
UTAH_UTES = "Utah Utes Men's"
TCU_HORNED_FROGS = "TCU Horned Frogs Men's"
GEORGIA_STATE_PANTHERS = "Georgia State Panthers Men's"
TEXAS_STATE_BOBCATS = "Texas State Bobcats Men's"
CINCINNATI_BEARCATS = "Cincinnati Bearcats Men's"
COLORADO_BUFFALOES = "Colorado Buffaloes Men's"
OMAHA_MAVERICKS = "Omaha Mavericks Men's"
DENVER_PIONEERS = "Denver Pioneers Men's"
PITT_PANTHERS = "Pitt Panthers Men's"
FLORIDA_STATE_SEMINOLES = "Florida State Seminoles Men's"
SMU_MUSTANGS = "SMU Mustangs Men's"
VIRGINIA_CAVALIERS = "Virginia Cavaliers Men's"
TEXAS_LONGHORNS = "Texas Longhorns Men's"
OKLAHOMA_SOONERS = "Oklahoma Sooners Men's"
MISSISSIPPI_STATE_BULLDOGS = "Mississippi State Bulldogs Men's"
AUBURN_TIGERS = "Auburn Tigers Men's"
MIAMI_FL_HURRIANCES = "Miami (FL) Hurricanes Men's"
DUKE_BLUE_DEVILS = "Duke Blue Devils Men's"
OLE_MISS_REBELS = "Ole Miss Rebels Men's"
ALABAMA_CRIMSON_TIDE = "Alabama Crimson Tide Men's"
MISSOURI_TIGERS = "Missouri Tigers Men's"
FLORIDA_GATORS = "Florida Gators Men's"
MARQUETTE_GOLDEN_EAGLES = "Marquette Golden Eagles Men's"
TEXAS_AM_AGGIES = "Texas A&M Aggies Men's"
KENTUCKY_WILDCATS = "Kentucky Wildcats Men's"
ILLINOIS_FIGHTING_ILLINI = "Illinois Fighting Illini Men's"
WISCONSIN_BADGERS = "Wisconsin Badgers Men's"
BAYLOR_BEARS = "Baylor Bears Men's"
ARIZONA_WILDCATS = "Arizona Wildcats Men's"
WESTERN_MICHIGAN_BRONCOS = "Western Michigan Broncos Men's"
KENT_STATE_GOLDEN_FLASHES = "Kent State Golden Flashes Men's"
ST_JOHNS_NY_RED_STORM = "St. John's (NY) Red Storm Men's"
SAINT_LOUIS_BILLIKENS = "Saint Louis Billikens Men's"
VALPARAISO_BEACONS = "Valparaiso Beacons Men's"
BELMONT_BRUINS = "Belmont Bruins Men's"
TEXAS_TECH_RED_RAIDERS = "Texas Tech Red Raiders Men's"
KANSAS_STATE_WILDCATS = "Kansas State Wildcats Men's"
MIAMI_OH_REDHAWKS = "Miami (OH) RedHawks Men's"
NORTHERN_ILLINOIS_HUSKIES = "Northern Illinois Huskies Men's"
EVANSVILLE_PURPLE_ACES = "Evansville Purple Aces Men's"
NORTHERN_IOWA_PANTHERS = "Northern Iowa Panthers Men's"
UCF_KNIGHTS = "UCF Knights Men's"
ARIZONA_STATE_SUN_DEVILS = "Arizona State Sun Devils Men's"
NEVADA_WOLF_PACK = "Nevada Wolf Pack Men's"
SAN_JOSE_STATE_SPARTANS = "San Jose State Spartans Men's"
USC_TROJANS = "USC Trojans Men's"
XAVIER_MUSKETEERS = "Xavier Musketeers Men's"
TOLEDO_ROCKETS = "Toledo Rockets Men's"
BUFFALO_BULLS = "Buffalo Bulls Men's"
BOWLING_GREEN_STATE_FALCONS = "Bowling Green State Falcons Men's"
EASTERN_MICHIGAN_EAGLES = "Eastern Michigan Eagles Men's"
CENTRAL_MICHIGAN_CHIPPEWAS = "Central Michigan Chippewas Men's"
NORTH_TEXAS_MEAN_GREEN = "North Texas Mean Green Men's"
EAST_CAROLINA_PIRATES = "East Carolina Pirates Men's"
BALL_STATE_CARDINALS = "Ball State Cardinals Men's"
LOUISVILLE_CARDINALS = "Louisville Cardinals Men's"
SYRACUSE_ORANGE = "Syracuse Orange Men's"
CHARLOTTE_49ERS = "Charlotte 49ers Men's"
WICHITA_STATE_SHOCKERS = "Wichita State Shockers Men's"
CREIGHTON_BLUEJAYS = "Creighton Bluejays Men's"
UTSA_ROADRUNNERS = "UTSA Roadrunners Men's"
RICE_OWLS = "Rice Owls Men's"
LITTLE_ROCK_TROJANS = "Little Rock Trojans Men's"
WYOMING_COWBOYS = "Wyoming Cowboys Men's"
OKLAHOMA_STATE_COWBOYS = "Oklahoma State Cowboys Men's"
BYU_COUGARS = "BYU Cougars Men's"
CLEMSON_TIGERS = "Clemson Tigers Men's"
GEORGIA_TECH_YELLOW_JACKETS = "Georgia Tech Yellow Jackets Men's"
ARKANSAS_RAZORBACKS = "Arkansas Razorbacks Men's"
LSU_FIGHTING_TIGERS = "LSU Fighting Tigers Men's"
COLORADO_STATE_RAMS = "Colorado State Rams Men's"
SAN_DIEGO_STATE_AZTECS = "San Diego State Aztecs Men's"
AMERICAN_EAGLES = "American Eagles Men's"
LOYOLA_MD_GREYHOUNDS = "Loyola (MD) Greyhounds Men's"
ALABAMA_STATE_HORNETS = "Alabama State Hornets Men's"
ALCORN_STATE_BRAVES = "Alcorn State Braves Men's"
PRAIRIE_VIEW_PANTHERS = "Prairie View Panthers Men's"
ARKANSAS_PINE_BLUFF_GOLDEN_LIONS = "Arkansas-Pine Bluff Golden Lions Men's"
FLORIDA_AM_RATTLERS = "Florida A&M Rattlers Men's"
GRAMBLING_TIGERS = "Grambling Tigers Men's"
NICHOLLS_STATE_COLONELS = "Nicholls State Colonels Men's"
INCARNATE_WORD_CARDINALS = "Incarnate Word Cardinals Men's"
ALABAMA_AM_BULLDOGS = "Alabama A&M Bulldogs Men's"
JACKSON_STATE_TIGERS = "Jackson State Tigers Men's"
TEXAS_SOUTHERN_TIGERS = "Texas Southern Tigers Men's"
MISSISSIPPI_VALLEY_STATE_DELTA_DEVILS = "Mississippi Valley State Delta Devils Men's"
NORFOLK_STATE_SPARTANS = "Norfolk State Spartans Men's"
MORGAN_STATE_BEARS = "Morgan State Bears Men's"
DELAWARE_STATE_HORNETS = "Delaware State Hornets Men's"
NORTH_CAROLINA_CENTRAL_EAGLES = "North Carolina Central Eagles Men's"
BOSTON_COLLEGE_EAGLES = "Boston College Eagles Men's"
NOTRE_DAME_FIGHTING_IRISH = "Notre Dame Fighting Irish Men's"
MARYLAND_EASTERN_SHORE_HAWKS = "Maryland-Eastern Shore Hawks Men's"
SOUTH_CAROLINA_STATE_BULLDOGS = "South Carolina State Bulldogs Men's"
SOUTHEASTERN_LOUISIANA_LIONS = "Southeastern Louisiana Lions Men's"
STEPHEN_F_AUSTIN_LUMBERJACKS = "Stephen F. Austin Lumberjacks Men's"
EAST_TEXAS_AM_LIONS = "East Texas A&M Lions Men's"
TEXAS_RIO_GRANDE_VALLEY_VAQUEROS = "Texas-Rio Grande Valley Vaqueros Men's"
MCNEESE_STATE_COWBOYS = "McNeese State Cowboys Men's"
HOUSTON_CHRISTIAN_HUSKIES = "Houston Christian Huskies Men's"
NEW_ORLEANS_PRIVATEERS = "New Orleans Privateers Men's"
LAMAR_CARDINALS = "Lamar Cardinals Men's"
BETHUNE_COOKMAN_WILDCATS = "Bethune-Cookman Wildcats Men's"
SOUTHERN_JAGUARS = "Southern Jaguars Men's"
NORTHWESTERN_STATE_DEMONS = "Northwestern State Demons Men's"
TEXAS_AM_CORPUS_CHRISTI_ISLANDERS = "Texas A&M-Corpus Christi Islanders Men's"
OREGON_DUCKS = "Oregon Ducks Men's"
MANHATTAN_JASPERS = "Manhattan Jaspers Men's"
LIU_SHARKS = "LIU Sharks Men's"
CENTRAL_CONNECTICUT_BLUE_DEVILS = "Central Connecticut Blue Devils Men's"
SAINT_FRANCIS_PA_RED_FLASH = "Saint Francis (PA) Red Flash Men's"
FDU_KNIGHTS = "FDU Knights Men's"
STONEHILL_SKYHAWKS = "Stonehill Skyhawks Men's"
CHICAGO_STATE_COUGARS = "Chicago State Cougars Men's"
MERCYHURST_LAKERS = "Mercyhurst Lakers Men's"
LE_MOYNE_DOLPHINS = "Le Moyne Dolphins Men's"
TULSA_GOLDEN_HURRICANE = "Tulsa Golden Hurricane Men's"
UCONN_HUSKIES = "UConn Huskies Men's"
PENN_QUAKERS = "Penn Quakers Men's"
DARTMOUTH_BIG_GREEN = "Dartmouth Big Green Men's"
UMASS_LOWELL_RIVER_HAWKS = "UMass-Lowell River Hawks Men's"
WESTERN_KENTUCKY_HILLTOPPERS = "Western Kentucky Hilltoppers Men's"
SAINT_MARYS_GAELS = "Saint Mary's Gaels Men's"
CORNELL_BIG_RED = "Cornell Big Red Men's"
COLUMBIA_LIONS = "Columbia Lions Men's"
PRINCETON_TIGERS = "Princeton Tigers Men's"
HARVARD_CRIMSON = "Harvard Crimson Men's"
PRESBYTERIAN_BLUE_HOSE = "Presbyterian Blue Hose Men's"
BROWN_BEARS = "Brown Bears Men's"
YALE_BULLDOGS = "Yale Bulldogs Men's"
EASTERN_WASHINGTON_EAGLES = "Eastern Washington Eagles Men's"
IDAHO_VANDALS = "Idaho Vandals Men's"
MIDDLE_TENNESSEE_BLUE_RAIDERS = "Middle Tennessee Blue Raiders Men's"
DETROIT_MERCY_TITANS = "Detroit Mercy Titans Men's"
SOUTHERN_UTAH_THUNDERBIRDS = "Southern Utah Thunderbirds Men's"
ST_THOMAS_TOMMIES = "St. Thomas Tommies Men's"
CAL_STATE_BAKERSFIELD_ROADRUNNERS = "Cal State Bakersfield Roadrunners Men's"
WAGNER_SEAHAWKS = "Wagner Seahawks Men's"
NCAAB_TEAM_IDENTITY_MAP: dict[str, str] = {
    "DePaul Blue Demons Men's": DEPAUL_BLUE_DEMONS,
    "Georgetown Hoyas Men's": GEORGETOWN_HOYAS,
    "Cal State Northridge Matadors Men's": CAL_STATE_NORTHRIDGE_MATADORS,
    "Hawaii Rainbow Warriors Men's": HAWAII_RAINBOW_WARRIORS,
    "Indiana Hoosiers Men's": INDIANA_HOOSIERS,
    "Ohio State Buckeyes Men's": OHIO_STATE_BUCKEYES,
    "Ohio Bobcats Men's": OHIO_BOBCATS,
    "Akron Zips Men's": AKRON_ZIPS,
    "Robert Morris Colonials Men's": ROBERT_MORRIS_COLONIALS,
    "Green Bay Phoenix Men's": GREEN_BAY_PHOENIX,
    "VCU Rams Men's": VCU_RAMS,
    "St. Joseph's Hawks Men's": ST_JOSEPHS_HAWKS,
    "Providence Friars Men's": PROVIDENCE_FRIARS,
    "Villanova Wildcats Men's": VILLANOVA_WILDCATS,
    "Youngstown State Penguins Men's": YOUNGSTOWN_STATE_PENGUINS,
    "Milwaukee Panthers Men's": MILWAUKEE_PANTHERS,
    "Iowa Hawkeyes Men's": IOWA_HAWKEYES,
    "UCLA Bruins Men's": UCLA_BRUINS,
    "Air Force Falcons Men's": AIR_FORCE_FALCONS,
    "Fresno State Bulldogs Men's": FRESNO_STATE_BULLDOGS,
    "Boise State Broncos Men's": BOISE_STATE_BRONCOS,
    "New Mexico Lobos Men's": NEW_MEXICO_LOBOS,
    "Gonzaga Bulldogs Men's": GONZAGA_BULLDOGS,
    "Oregon State Beavers Men's": OREGON_STATE_BEAVERS,
    "Memphis Tigers Men's": MEMPHIS_TIGERS,
    "Temple Owls Men's": TEMPLE_OWLS,
    "Michigan Wolverines Men's": MICHIGAN_WOLVERINES,
    "Minnesota Golden Gophers Men's": MINNESOTA_GOLDEN_GOPHERS,
    "College of Charleston Cougars Men's": COLLEGE_CHARLESTON_COUGARS,
    "Campbell Fighting Camels Men's": CAMPBELL_FIGHTING_CAMELS,
    "Stony Brook Seawolves Men's": STONY_BROOKS_SEAWOLVES,
    "Delaware Fightin' Blue Hens Men's": DELAWARE_FIGHTIN_BLUE_HENS,
    "New Mexico State Aggies Men's": NEW_MEXICO_STATE_AGGIES,
    "Florida International Panthers Men's": FLORIDA_INTERNATIONAL_PANTHERS,
    "New Hampshire Wildcats Men's": NEW_HAMPSHIRE_WILDCATS,
    "UMBC Retrievers Men's": UMBC_RETRIEVERS,
    "Appalachian State Mountaineers Men's": APPALACHIAN_STATE_MOUNTAINEERS,
    "Old Dominion Monarchs Men's": OLD_DOMINION_MONARCHS,
    "Marist Red Foxes Men's": MARIST_RED_FOXES,
    "St. Peter's Peacocks Men's": ST_PETERS_PEACOCKS,
    "Queens (NC) Royals Men's": QUEENS_ROYALS,
    "Stetson Hatters Men's": STETSON_HATTERS,
    "Eastern Kentucky Colonels Men's": EASTERN_KENTUCKY_COLONELS,
    "Austin Peay Governors Men's": AUSTIN_PEAY_GOVERNORS,
    "Rutgers Scarlet Knights Men's": RUTGERS_SCARLET_KNIGHTS,
    "Nebraska Cornhuskers Men's": NEBRASKA_CORNHUSKERS,
    "Maryland Terrapins Men's": MARYLAND_TERRAPINS,
    "Northwestern Wildcats Men's": NORTHWESTERN_WILDCATS,
    "SIU-Edwardsville Cougars Men's": SIU_EDWARDSVILLE_COUGARS,
    "Tennessee State Tigers Men's": TENNESSEE_STATE_TIGERS,
    "Southeast Missouri State Redhawks Men's": SOUTHEAST_MISSOURI_STATE_REDHAWKS,
    "Tennessee Tech Golden Eagles Men's": TENNESSEE_TECH_GOLDEN_EAGLES,
    "UC-Riverside Highlanders Men's": UC_RIVERSIDE_HIGHLANDERS,
    "UCSB Gauchos Men's": UCSB_GAUCHOS,
    "Washington State Cougars Men's": WASHINGTON_STATE_COUGARS,
    "San Diego Toreros Men's": SAN_DIEGO_TOREROS,
    "Merrimack Warriors Men's": MERRIMACK_WARRIORS,
    "Quinnipiac Bobcats Men's": QUINNIPIAC_BOBCATS,
    "Coppin State Eagles Men's": COPPIN_STATE_EAGLES,
    "Maine Black Bears Men's": MAINE_BLACK_BEARS,
    "NJIT Highlanders Men's": NJIT_HIGHLANDERS,
    "Bryant Bulldogs Men's": BRYANT_BULLDOGS,
    "Albany (NY) Great Danes Men's": ALBANY_NY_GREAT_DANES,
    "Lipscomb Bisons Men's": LIPSCOMB_BISONS,
    "Bellarmine Knights Men's": BELLARMINE_KNIGHTS,
    "West Georgia Wolves Men's": WEST_GEORGIA_WOLVES,
    "Florida Gulf Coast Eagles Men's": FLORIDA_GULF_COAST_EAGLES,
    "Georgia Southern Eagles Men's": GEORGIA_SOUTHERN_EAGLES,
    "Coastal Carolina Chanticleers Men's": COASTAL_CAROLINA_CHANTICLEERS,
    "Elon Phoenix Men's": ELON_PHOENIX,
    "Drexel Dragons Men's": DREXEL_DRAGONS,
    "Niagara Purple Eagles Men's": NIAGARA_PURPLE_EAGLES,
    "Fairfield Stags Men's": FAIRFIELD_STAGS,
    "William & Mary Tribe Men's": WILLIAM_AND_MARY_TRIBE,
    "Hampton Pirates Men's": HAMPTON_PIRATES,
    "Canisius Golden Griffins Men's": CANISIUS_GOLDEN_GRIFFINS,
    "Iona Gaels Men's": IONA_GAELS,
    "North Alabama Lions Men's": NORTH_ALABAMA_LIONS,
    "Jacksonville Dolphins Men's": JACKSONVILLE_DOLPHINS,
    "Louisiana Tech Bulldogs Men's": LOUISIANA_TECH_BULLDOGS,
    "Jacksonville State Gamecocks Men's": JACKSONVILLE_STATE_GAMECOCKS,
    "Marshall Thundering Herd Men's": MARSHALL_THUNDERING_HERD,
    "James Madison Dukes Men's": JAMES_MADISON_DUKES,
    "Sam Houston Bearkats Men's": SAM_HOUSTON_BEARKATS,
    "Kennesaw State Owls Men's": KENNESAW_STATE_OWLS,
    "High Point Panthers Men's": HIGH_POINT_PANTHERS,
    "Longwood Lancers Men's": LONGWOOD_LANCERS,
    "Western Illinois Leathernecks Men's": WESTERN_ILLINOIS_LEATHERNECKS,
    "Morehead State Eagles Men's": MOREHEAD_STATE_EAGLES,
    "Monmouth Hawks Men's": MONMOUTH_HAWKS,
    "North Carolina A&T Aggies Men's": NORTH_CAROLINA_ATT_AGGIES,
    "Northeastern Huskies Men's": NORTHEASTERN_HUSKIES,
    "UNC Wilmington Seahawks Men's": UNC_WILMINGTON_SEAHAWKS,
    "Central Arkansas Bears Men's": CENTRAL_ARKANSAS_BEARS,
    "North Florida Ospreys Men's": NORTH_FLORIDA_OSPREYS,
    "Mount St. Mary's Mountaineers Men's": MOUNT_ST_MARYS_MOUNTAINEERS,
    "Rider Broncs Men's": RIDER_BRONCS,
    "Sacred Heart Pioneers Men's": SACRED_HEART_PIONEERS,
    "Siena Saints Men's": SIENA_SAINTS,
    "Hofstra Pride Men's": HOFSTRA_PRIDE,
    "Towson Tigers Men's": TOWSON_TIGERS,
    "Binghamton Bearcats Men's": BINGHAMTON_BEARCATS,
    "Vermont Catamounts Men's": VERMONT_CATAMOUNTS,
    "Louisiana Ragin' Cajuns Men's": LOUISIANA_RAGIN_CAJUNS,
    "Arkansas State Red Wolves Men's": ARKANSAS_STATE_RED_WOLVES,
    "Montana State Bobcats Men's": MONTANA_STATE_BOBCATS,
    "Idaho State Bengals Men's": IDAHO_STATE_BENGALS,
    "Kansas City Roos Men's": KANSAS_CITY_ROOS,
    "North Dakota State Bison Men's": NORTH_DAKOTA_STATE_BISON,
    "Sacramento State Hornets Men's": SACRAMENTO_STATE_HORNETS,
    "Northern Arizona Lumberjacks Men's": NORTHERN_ARIZONA_LUMBERJACKS,
    "Oral Roberts Golden Eagles Men's": ORAL_ROBERTS_GOLDEN_EAGLES,
    "South Dakota Coyotes Men's": SOUTH_DAKOTA_COYOTES,
    "North Dakota Fighting Hawks Men's": NORTH_DAKOTA_FIGHTING_HAWKS,
    "South Dakota State Jackrabbits Men's": SOUTH_DAKOTA_STATE_JACKRABBITS,
    "Lindenwood Lions Men's": LINDENWOOD_LIONS,
    "Southern Indiana Screaming Eagles Men's": SOUTHERN_INDIANA_SCREAMING_EAGLES,
    "California Baptist Lancers Men's": CALIFORNIA_BAPTIST_LANCERS,
    "Tarleton State Texans Men's": TARLETON_STATE_TEXANS,
    "Eastern Illinois Panthers Men's": EASTERN_ILLINOIS_PANTHERS,
    "UT-Martin Skyhawks Men's": UT_MARTIN_SKYHAWKS,
    "Utah Valley Wolverines Men's": UTAH_VALLEY_WOLVERINES,
    "UT Arlington Mavericks Men's": UT_ARLINGTON_MAVERICKS,
    "Abilene Christian Wildcats Men's": ABILENE_CHRISTIAN_WILDCATS,
    "Grand Canyon Antelopes Men's": GRAND_CANYON_ANTELOPES,
    "UTEP Miners Men's": UTEP_MINERS,
    "Liberty Flames Men's": LIBERTY_FLAMES,
    "Portland State Vikings Men's": PORTLAND_STATE_VIKINGS,
    "Northern Colorado Bears Men's": NORTHERN_COLORADO_BEARS,
    "Montana Grizzlies Men's": MONTANA_GRIZZLIES,
    "Weber State Wildcats Men's": WEBER_STATE_WILDCATS,
    "UC-Davis Aggies Men's": UC_DAVIS_AGGIES,
    "Cal Poly Mustangs Men's": CAL_POLY_MUSTANGS,
    "Cal State Fullerton Titans Men's": CAL_STATE_FULLERTON_TITANS,
    "UC-Irvine Anteaters Men's": UC_IRVINE_ANTEATERS,
    "UC-San Diego Tritons Men's": UC_SAN_DIEGO_TRITONS,
    "Long Beach State Beach Men's": LONG_BEACH_STATE_BEACH,
    "Santa Clara Broncos Men's": SANTA_CLARA_BRONCOS,
    "Loyola Marymount Lions Men's": LOYOLA_MARYMOUNT_LIONS,
    "San Francisco Dons Men's": SAN_FRANCISCO_DONS,
    "Pepperdine Waves Men's": PEPPERDINE_WAVES,
    "Pacific Tigers Men's": PACIFIC_TIGERS,
    "Portland Pilots Men's": PORTLAND_PILOTS,
    "Utah Tech Trailblazers Men's": UTAH_TECH_TRAILBLAZERS,
    "Seattle Redhawks Men's": SEATTLE_REDHAWKS,
    "Kansas Jayhawks Men's": KANSAS_JAYHAWKS,
    "Iowa State Cyclones Men's": IOWA_STATE_CYCLONES,
    "Georgia Bulldogs Men's": GEORGIA_BULLDOGS,
    "Tennessee Volunteers Men's": TENNESSEE_VOLUNTEERS,
    "West Virginia Mountaineers Men's": WEST_VIRGINIA_MOUNTAINEERS,
    "Houston Cougars Men's": HOUSTON_COUGARS,
    "Penn State Nittany Lions Men's": PENN_STATE_NITTANY_LIONS,
    "Michigan State Spartans Men's": MICHIGAN_STATE_SPARTANS,
    "Purdue Boilermakers Men's": PURDUE_BOILERMAKERS,
    "Washington Huskies Men's": WASHINGTON_HUSKIES,
    "Utah State Aggies Men's": UTAH_STATE_AGGIES,
    "UNLV Rebels Men's": UNLV_REBELS,
    "Lafayette Leopards Men's": LAFAYETTE_LEOPARDS,
    "Army Black Knights Men's": ARMY_BLACK_KNIGHTS,
    "Wofford Terriers Men's": WOFFORD_TERRIERS,
    "Chattanooga Mocs Men's": CHATTANOOGA_MOCS,
    "George Mason Patriots Men's": GEORGE_MASON_PATRIOTS,
    "Dayton Flyers Men's": DAYTON_FLYERS,
    "Davidson Wildcats Men's": DAVIDSON_WILDCATS,
    "La Salle Explorers Men's": LA_SALLE_EXPLORERS,
    "Samford Bulldogs Men's": SAMFORD_BULLDOGS,
    "Mercer Bears Men's": MERCER_BEARS,
    "Gardner-Webb Runnin' Bulldogs Men's": GARDNER_WEBB_RUNNIN_BULLDOGS,
    "Radford Highlanders Men's": RADFORD_HIGHLANDERS,
    "Illinois State Redbirds Men's": ILLINOIS_STATE_REDBIRDS,
    "Drake Bulldogs Men's": DRAKE_BULLDOGS,
    "Rhode Island Rams Men's": RHODE_ISLAND_RAMS,
    "Loyola (IL) Ramblers Men's": LOYOLA_IL_RAMBLERS,
    "Southern Illinois Salukis Men's": SOUTHERN_ILLINOIS_SALUKIS,
    "Missouri State Bears Men's": MISSOURI_STATE_BEARS,
    "Lehigh Mountain Hawks Men's": LEHIGH_MOUNTAIN_HAWKS,
    "Boston University Terriers Men's": BOSTON_UNIVERSITY_TERRIERS,
    "USC Upstate Spartans Men's": USC_UPSTATE_SPARTANS,
    "UNC Asheville Bulldogs Men's": UNC_ASHEVILLE_BULLDOGS,
    "South Carolina Gamecocks Men's": SOUTH_CAROLINA_GAMECOCKS,
    "Vanderbilt Commodores Men's": VANDERBILT_COMMODORES,
    "Western Carolina Catamounts Men's": WESTERN_CAROLINA_CATAMOUNTS,
    "VMI Keydets Men's": VMI_KEYDETS,
    "Stanford Cardinal Men's": STANFORD_CARDINAL,
    "Wake Forest Demon Deacons Men's": WAKE_FOREST_DEMON_DEACONS,
    "Charleston Southern Buccaneers Men's": CHARLESTON_SOUTHERN_BUCCANEERS,
    "Winthrop Eagles Men's": WINTHROP_EAGLES,
    "Holy Cross Crusaders Men's": HOLY_CROSS_CRUSADERS,
    "Bucknell Bison Men's": BUCKNELL_BISON,
    "Seton Hall Pirates Men's": SETON_HALL_PIRATES,
    "Butler Bulldogs Men's": BUTLER_BULLDOGS,
    "Northern Kentucky Norse Men's": NORTHERN_KENTUCKY_NORSE,
    "Cleveland State Vikings Men's": CLEVELAND_STATE_VIKINGS,
    "UMass Minutemen Men's": UMASS_MINUTEMEN,
    "Fordham Rams Men's": FORDHAM_RAMS,
    "ETSU Buccaneers Men's": ETSU_BUCCANEERS,
    "Furman Paladins Men's": FURMANS_PALADINS,
    "Duquesne Dukes Men's": DUQUESNE_DUKES,
    "George Washington Revolutionaries Men's": GEORGE_WASHINGTON_REVOLUTIONARIES,
    "Howard Bison Men's": HOWARD_BISON,
    "Troy Trojans Men's": TROY_TROJANS,
    "Louisiana-Monroe Warhawks Men's": LOUISIANA_MONRE_WARHAWKS,
    "Colgate Raiders Men's": COLGATE_RAIDERS,
    "Navy Midshipmen Men's": NAVY_MIDSHIPMEN,
    "California Golden Bears Men's": CALIFORNIA_GOLDEN_BEARS,
    "UNC Tar Heels Men's": UNC_TAR_HEELS,
    "The Citadel Bulldogs Men's": THE_CITADEL_BULLDOGS,
    "UNC Greensboro Spartans Men's": UNC_GREENSBORO_SPARTANS,
    "IU Indianapolis Jaguars Men's": IU_INDIANAPOLIS_JAGUARS,
    "Oakland Golden Grizzlies Men's": OAKLAND_GOLDEN_GRIZZLIES,
    "UAB Blazers Men's": UAB_BLAZERS,
    "South Florida Bulls Men's": SOUTH_FLORIDA_BULLS,
    "Richmond Spiders Men's": RICHMOND_SPIDERS,
    "St. Bonaventure Bonnies Men's": ST_BONAVENTURE_BONNIES,
    "Florida Atlantic Owls Men's": FLORIDA_ATLANTIC_OWLS,
    "Tulane Green Wave Men's": TULANE_GREEN_WAVE,
    "NC State Wolfpack Men's": NC_STATE_WOLFPACK,
    "Virginia Tech Hokies Men's": VIRGINIA_TECH_HOKIES,
    "Purdue Fort Wayne Mastodons Men's": PURDUE_FORT_WAYNE_MASTODONS,
    "Wright State Raiders Men's": WRIGHT_STATE_RAIDERS,
    "Indiana State Sycamores Men's": INDIANA_STATE_SYCAMORES,
    "Bradley Braves Men's": BRADLEY_BRAVES,
    "Murray State Racers Men's": MURRAY_STATE_RACERS,
    "UIC Flames Men's": UIC_FLAMES,
    "Southern Miss Golden Eagles Men's": SOUTHERN_MISS_GOLDEN_EAGLES,
    "South Alabama Jaguars Men's": SOUTH_ALABAMA_JAGUARS,
    "Utah Utes Men's": UTAH_UTES,
    "TCU Horned Frogs Men's": TCU_HORNED_FROGS,
    "Georgia State Panthers Men's": GEORGIA_STATE_PANTHERS,
    "Texas State Bobcats Men's": TEXAS_STATE_BOBCATS,
    "Cincinnati Bearcats Men's": CINCINNATI_BEARCATS,
    "Colorado Buffaloes Men's": COLORADO_BUFFALOES,
    "Omaha Mavericks Men's": OMAHA_MAVERICKS,
    "Denver Pioneers Men's": DENVER_PIONEERS,
    "Pitt Panthers Men's": PITT_PANTHERS,
    "Florida State Seminoles Men's": FLORIDA_STATE_SEMINOLES,
    "SMU Mustangs Men's": SMU_MUSTANGS,
    "Virginia Cavaliers Men's": VIRGINIA_CAVALIERS,
    "Texas Longhorns Men's": TEXAS_LONGHORNS,
    "Oklahoma Sooners Men's": OKLAHOMA_SOONERS,
    "Mississippi State Bulldogs Men's": MISSISSIPPI_STATE_BULLDOGS,
    "Auburn Tigers Men's": AUBURN_TIGERS,
    "Miami (FL) Hurricanes Men's": MIAMI_FL_HURRIANCES,
    "Duke Blue Devils Men's": DUKE_BLUE_DEVILS,
    "Ole Miss Rebels Men's": OLE_MISS_REBELS,
    "Alabama Crimson Tide Men's": ALABAMA_CRIMSON_TIDE,
    "Missouri Tigers Men's": MISSOURI_TIGERS,
    "Florida Gators Men's": FLORIDA_GATORS,
    "Marquette Golden Eagles Men's": MARQUETTE_GOLDEN_EAGLES,
    "Texas A&M Aggies Men's": TEXAS_AM_AGGIES,
    "Kentucky Wildcats Men's": KENTUCKY_WILDCATS,
    "Illinois Fighting Illini Men's": ILLINOIS_FIGHTING_ILLINI,
    "Wisconsin Badgers Men's": WISCONSIN_BADGERS,
    "Baylor Bears Men's": BAYLOR_BEARS,
    "Arizona Wildcats Men's": ARIZONA_WILDCATS,
    "Western Michigan Broncos Men's": WESTERN_MICHIGAN_BRONCOS,
    "Kent State Golden Flashes Men's": KENT_STATE_GOLDEN_FLASHES,
    "St. John's (NY) Red Storm Men's": ST_JOHNS_NY_RED_STORM,
    "Saint Louis Billikens Men's": SAINT_LOUIS_BILLIKENS,
    "Valparaiso Beacons Men's": VALPARAISO_BEACONS,
    "Belmont Bruins Men's": BELMONT_BRUINS,
    "Texas Tech Red Raiders Men's": TEXAS_TECH_RED_RAIDERS,
    "Kansas State Wildcats Men's": KANSAS_STATE_WILDCATS,
    "Miami (OH) RedHawks Men's": MIAMI_OH_REDHAWKS,
    "Northern Illinois Huskies Men's": NORTHERN_ILLINOIS_HUSKIES,
    "Evansville Purple Aces Men's": EVANSVILLE_PURPLE_ACES,
    "Northern Iowa Panthers Men's": NORTHERN_IOWA_PANTHERS,
    "UCF Knights Men's": UCF_KNIGHTS,
    "Arizona State Sun Devils Men's": ARIZONA_STATE_SUN_DEVILS,
    "Nevada Wolf Pack Men's": NEVADA_WOLF_PACK,
    "San Jose State Spartans Men's": SAN_JOSE_STATE_SPARTANS,
    "USC Trojans Men's": USC_TROJANS,
    "Xavier Musketeers Men's": XAVIER_MUSKETEERS,
    "Toledo Rockets Men's": TOLEDO_ROCKETS,
    "Buffalo Bulls Men's": BUFFALO_BULLS,
    "Bowling Green State Falcons Men's": BOWLING_GREEN_STATE_FALCONS,
    "Eastern Michigan Eagles Men's": EASTERN_MICHIGAN_EAGLES,
    "Central Michigan Chippewas Men's": CENTRAL_MICHIGAN_CHIPPEWAS,
    "North Texas Mean Green Men's": NORTH_TEXAS_MEAN_GREEN,
    "East Carolina Pirates Men's": EAST_CAROLINA_PIRATES,
    "Ball State Cardinals Men's": BALL_STATE_CARDINALS,
    "Louisville Cardinals Men's": LOUISVILLE_CARDINALS,
    "Syracuse Orange Men's": SYRACUSE_ORANGE,
    "Charlotte 49ers Men's": CHARLOTTE_49ERS,
    "Wichita State Shockers Men's": WICHITA_STATE_SHOCKERS,
    "Creighton Bluejays Men's": CREIGHTON_BLUEJAYS,
    "UTSA Roadrunners Men's": UTSA_ROADRUNNERS,
    "Rice Owls Men's": RICE_OWLS,
    "Little Rock Trojans Men's": LITTLE_ROCK_TROJANS,
    "Wyoming Cowboys Men's": WYOMING_COWBOYS,
    "Oklahoma State Cowboys Men's": OKLAHOMA_STATE_COWBOYS,
    "BYU Cougars Men's": BYU_COUGARS,
    "Clemson Tigers Men's": CLEMSON_TIGERS,
    "Georgia Tech Yellow Jackets Men's": GEORGIA_TECH_YELLOW_JACKETS,
    "Arkansas Razorbacks Men's": ARKANSAS_RAZORBACKS,
    "LSU Fighting Tigers Men's": LSU_FIGHTING_TIGERS,
    "Colorado State Rams Men's": COLORADO_STATE_RAMS,
    "San Diego State Aztecs Men's": SAN_DIEGO_STATE_AZTECS,
    "American Eagles Men's": AMERICAN_EAGLES,
    "Loyola (MD) Greyhounds Men's": LOYOLA_MD_GREYHOUNDS,
    "Alabama State Hornets Men's": ALABAMA_STATE_HORNETS,
    "Alcorn State Braves Men's": ALCORN_STATE_BRAVES,
    "Prairie View Panthers Men's": PRAIRIE_VIEW_PANTHERS,
    "Arkansas-Pine Bluff Golden Lions Men's": ARKANSAS_PINE_BLUFF_GOLDEN_LIONS,
    "Florida A&M Rattlers Men's": FLORIDA_AM_RATTLERS,
    "Grambling Tigers Men's": GRAMBLING_TIGERS,
    "Nicholls State Colonels Men's": NICHOLLS_STATE_COLONELS,
    "Incarnate Word Cardinals Men's": INCARNATE_WORD_CARDINALS,
    "Alabama A&M Bulldogs Men's": ALABAMA_AM_BULLDOGS,
    "Jackson State Tigers Men's": JACKSON_STATE_TIGERS,
    "Texas Southern Tigers Men's": TEXAS_SOUTHERN_TIGERS,
    "Mississippi Valley State Delta Devils Men's": MISSISSIPPI_VALLEY_STATE_DELTA_DEVILS,
    "Norfolk State Spartans Men's": NORFOLK_STATE_SPARTANS,
    "Morgan State Bears Men's": MORGAN_STATE_BEARS,
    "Delaware State Hornets Men's": DELAWARE_STATE_HORNETS,
    "North Carolina Central Eagles Men's": NORTH_CAROLINA_CENTRAL_EAGLES,
    "Boston College Eagles Men's": BOSTON_COLLEGE_EAGLES,
    "Notre Dame Fighting Irish Men's": NOTRE_DAME_FIGHTING_IRISH,
    "Maryland-Eastern Shore Hawks Men's": MARYLAND_EASTERN_SHORE_HAWKS,
    "South Carolina State Bulldogs Men's": SOUTH_CAROLINA_STATE_BULLDOGS,
    "Southeastern Louisiana Lions Men's": SOUTHEASTERN_LOUISIANA_LIONS,
    "Stephen F. Austin Lumberjacks Men's": STEPHEN_F_AUSTIN_LUMBERJACKS,
    "East Texas A&M Lions Men's": EAST_TEXAS_AM_LIONS,
    "Texas-Rio Grande Valley Vaqueros Men's": TEXAS_RIO_GRANDE_VALLEY_VAQUEROS,
    "McNeese State Cowboys Men's": MCNEESE_STATE_COWBOYS,
    "Houston Christian Huskies Men's": HOUSTON_CHRISTIAN_HUSKIES,
    "New Orleans Privateers Men's": NEW_ORLEANS_PRIVATEERS,
    "Lamar Cardinals Men's": LAMAR_CARDINALS,
    "Bethune-Cookman Wildcats Men's": BETHUNE_COOKMAN_WILDCATS,
    "Southern Jaguars Men's": SOUTHERN_JAGUARS,
    "Northwestern State Demons Men's": NORTHWESTERN_STATE_DEMONS,
    "Texas A&M-Corpus Christi Islanders Men's": TEXAS_AM_CORPUS_CHRISTI_ISLANDERS,
    "Oregon Ducks Men's": OREGON_DUCKS,
    "Manhattan Jaspers Men's": MANHATTAN_JASPERS,
    "LIU Sharks Men's": LIU_SHARKS,
    "Central Connecticut Blue Devils Men's": CENTRAL_CONNECTICUT_BLUE_DEVILS,
    "Saint Francis (PA) Red Flash Men's": SAINT_FRANCIS_PA_RED_FLASH,
    "FDU Knights Men's": FDU_KNIGHTS,
    "Stonehill Skyhawks Men's": STONEHILL_SKYHAWKS,
    "Chicago State Cougars Men's": CHICAGO_STATE_COUGARS,
    "Mercyhurst Lakers Men's": MERCYHURST_LAKERS,
    "Le Moyne Dolphins Men's": LE_MOYNE_DOLPHINS,
    "Tulsa Golden Hurricane Men's": TULSA_GOLDEN_HURRICANE,
    "UConn Huskies Men's": UCONN_HUSKIES,
    "Penn Quakers Men's": PENN_QUAKERS,
    "Dartmouth Big Green Men's": DARTMOUTH_BIG_GREEN,
    "UMass-Lowell River Hawks Men's": UMASS_LOWELL_RIVER_HAWKS,
    "Western Kentucky Hilltoppers Men's": WESTERN_KENTUCKY_HILLTOPPERS,
    "Saint Mary's Gaels Men's": SAINT_MARYS_GAELS,
    "Cornell Big Red Men's": CORNELL_BIG_RED,
    "Columbia Lions Men's": COLUMBIA_LIONS,
    "Princeton Tigers Men's": PRINCETON_TIGERS,
    "Harvard Crimson Men's": HARVARD_CRIMSON,
    "Presbyterian Blue Hose Men's": PRESBYTERIAN_BLUE_HOSE,
    "Brown Bears Men's": BROWN_BEARS,
    "Yale Bulldogs Men's": YALE_BULLDOGS,
    "Eastern Washington Eagles Men's": EASTERN_WASHINGTON_EAGLES,
    "Idaho Vandals Men's": IDAHO_VANDALS,
    "Middle Tennessee Blue Raiders Men's": MIDDLE_TENNESSEE_BLUE_RAIDERS,
    "Detroit Mercy Titans Men's": DETROIT_MERCY_TITANS,
    "Southern Utah Thunderbirds Men's": SOUTHERN_UTAH_THUNDERBIRDS,
    "St. Thomas Tommies Men's": ST_THOMAS_TOMMIES,
    "Cal State Bakersfield Roadrunners Men's": CAL_STATE_BAKERSFIELD_ROADRUNNERS,
    "Wagner Seahawks Men's": WAGNER_SEAHAWKS,
}
NCAAB_VENUE_IDENTITY_MAP: dict[str, str] = {}
NCAAB_PLAYER_IDENTITY_MAP: dict[str, str] = {}


class NCAABCombinedLeagueModel(CombinedLeagueModel):
    """NCAAB combined implementation of the league model."""

    def __init__(self, session: ProxySession, league_filter: str | None) -> None:
        super().__init__(
            session,
            League.NCAAB,
            [
                NCAABSportsReferenceLeagueModel(session, position=0),
                NCAABESPNLeagueModel(session, position=1),
                NCAABOddsPortalLeagueModel(session, position=2),
            ],
            league_filter,
        )

    @classmethod
    def team_identity_map(cls) -> dict[str, str]:
        return NCAAB_TEAM_IDENTITY_MAP

    @classmethod
    def venue_identity_map(cls) -> dict[str, str]:
        return NCAAB_VENUE_IDENTITY_MAP

    @classmethod
    def player_identity_map(cls) -> dict[str, str]:
        return NCAAB_PLAYER_IDENTITY_MAP

    @classmethod
    def name(cls) -> str:
        return "ncaab-combined-league-model"



"""Tests for the ESPN player model class."""
import datetime
import os
import unittest

import requests_mock
import requests_cache
from sportsball.data.espn.espn_player_model import create_espn_player_model


class TestESPNPlayerModel(unittest.TestCase):

    def setUp(self):
        self._session = requests_cache.CachedSession(backend="memory")
        self.dir = os.path.dirname(__file__)

    def test_identifier(self):
        dt = datetime.datetime(2023, 9, 15, 0, 15)
        identifier = "a"
        statistics_url = "http://sports.core.api.espn.com/v2/sports/football/leagues/nfl/events/401671855/competitions/401671855/competitors/5/roster/16837/statistics/0?lang=en&region=us"
        athletes_url = "http://sports.core.api.espn.com/v2/sports/football/leagues/nfl/seasons/2024/athletes/16837?lang=en&region=us"
        position_url = "http://sports.core.api.espn.com/v2/sports/football/leagues/nfl/positions/32?lang=en&region=us"
        with requests_mock.Mocker() as m:
            with open(os.path.join(self.dir, "16837_statistics.json"), "rb") as f:
                m.get(statistics_url, content=f.read())
            with open(os.path.join(self.dir, "16837_athletes.json"), "rb") as f:
                m.get(athletes_url, content=f.read())
            with open(os.path.join(self.dir, "32_positions.json"), "rb") as f:
                m.get(position_url, content=f.read())
            player_model = create_espn_player_model(
                session=self._session,
                player={
                    "playerId": identifier,
                    "period": 0,
                    "active": False,
                    "starter": True,
                    "forPlayerId": 0,
                    "jersey": "93",
                    "valid": True,
                    "athlete": {
                        "$ref": athletes_url,
                    },
                    "position": {
                        "$ref": position_url,
                    },
                    "statistics": {
                        "$ref": statistics_url,
                    },
                    "didNotPlay": False,
                    "displayName": "S. Harris",
                },
                dt=dt,
                positions_validator={"DT": "DT"},
            )

            self.assertEqual(player_model.identifier, identifier)



"""The enumeration of the different supported positions."""

from enum import StrEnum


class Position(StrEnum):
    """An enumeration over the different positions."""

    POWER_FORWARD = "PF"


_POSITIONS = {str(x): x for x in Position}


def position_from_str(position_str: str) -> Position:
    """Find a position from a string."""
    position = _POSITIONS.get(position_str)
    if position is None:
        raise ValueError(f"Unrecognised position: {position_str}")
    return position



"""Tests for the google news model class."""
import datetime
import unittest

import requests_cache
from sportsball.data.google.google_news_model import create_google_news_models
from sportsball.data.league import League


class TestGoogleNewsModel(unittest.TestCase):

    def setUp(self):
        self.session = requests_cache.CachedSession(backend="memory")

    def test_dt_late(self):
        dt = datetime.datetime.now() + datetime.timedelta(days=1000)
        news_model = create_google_news_models("Imperial Arena at Atlantis Resort, Nassau", self.session, dt, League.NFL)
        self.assertListEqual(news_model, [])



"""The HKJC league model."""

# pylint: disable=line-too-long
import datetime
import urllib.parse
from typing import Iterator
from urllib.parse import urlparse

import tqdm
from bs4 import BeautifulSoup

from ....proxy_session import ProxySession
from ...game_model import GameModel
from ...league import League
from ...league_model import LeagueModel
from .hkjc_hkjc_game_model import (RACE_COURSE_QUERY_KEY, RACE_DATE_QUERY_KEY,
                                   RACE_NUMBER_QUERY_KEY,
                                   create_hkjc_hkjc_game_model)
from .hkjc_hkjc_venue_model import HAPPY_VALLEY_VENUE_CODE, SHA_TIN_VENUE_CODE


class HKJCHKJCLeagueModel(LeagueModel):
    """HKJC HKJC implementation of the league model."""

    def __init__(self, session: ProxySession, position: int | None = None) -> None:
        super().__init__(League.HKJC, session, position=position)

    @classmethod
    def name(cls) -> str:
        return "hkjc-league-model"

    @property
    def games(self) -> Iterator[GameModel]:  # type: ignore
        def games_for_date(dt: datetime.date) -> Iterator[GameModel]:
            dt_str = dt.strftime("%Y/%m/%d")
            for course in [SHA_TIN_VENUE_CODE, HAPPY_VALLEY_VENUE_CODE]:
                with self.session.wayback_disabled():
                    response = self.session.get(
                        f"https://racing.hkjc.com/racing/information/English/Racing/LocalResults.aspx?{RACE_DATE_QUERY_KEY}={dt_str}&{RACE_COURSE_QUERY_KEY}={course}&{RACE_NUMBER_QUERY_KEY}=1",
                    )
                response.raise_for_status()
                game_model = create_hkjc_hkjc_game_model(
                    self.session, response.text, response.url
                )
                if game_model is not None:
                    yield game_model
                soup = BeautifulSoup(response.text, "lxml")
                seen_urls = {response.url}
                for a in soup.find_all("a", href=True):
                    url = urllib.parse.urljoin(response.url, a.get("href"))
                    if url in seen_urls:
                        continue
                    o = urlparse(url)
                    if not o.path.endswith("/LocalResults.aspx"):
                        continue
                    query = urllib.parse.parse_qs(o.query)
                    if RACE_NUMBER_QUERY_KEY not in query:
                        continue
                    if query[RACE_COURSE_QUERY_KEY][0] != course:
                        continue
                    with self.session.wayback_disabled():
                        response = self.session.get(url)
                    response.raise_for_status()
                    game_model = create_hkjc_hkjc_game_model(
                        self.session, response.text, response.url
                    )
                    if game_model is not None:
                        yield game_model
                    seen_urls.add(url)

        today_date = datetime.datetime.today().date()
        last_scraped_date: datetime.date | None = None
        current_date = today_date

        with tqdm.tqdm(position=self.position) as pbar:
            while (
                last_scraped_date is None
                or (last_scraped_date - current_date).days < 360
            ):
                if current_date >= today_date:
                    with self.session.cache_disabled():
                        for game_model in games_for_date(current_date):
                            last_scraped_date = current_date
                            pbar.update(1)
                            yield game_model
                else:
                    for game_model in games_for_date(current_date):
                        last_scraped_date = current_date
                        pbar.update(1)
                        yield game_model
                pbar.set_description(f"HKJC {current_date}")
                current_date -= datetime.timedelta(days=1)



"""AFL OddsPortal league model."""

# pylint: disable=line-too-long

from ....proxy_session import ProxySession
from ...league import League
from ...oddsportal.oddsportal_league_model import OddsPortalLeagueModel


class AFLOddsPortalLeagueModel(OddsPortalLeagueModel):
    """AFL OddsPortal implementation of the league model."""

    def __init__(self, session: ProxySession, position: int | None = None) -> None:
        super().__init__(League.AFL, session, position=position)

    @classmethod
    def name(cls) -> str:
        return "afl-oddsportal-league-model"



"""The enumeration of the different supported bets."""

# pylint: disable=too-many-return-statements,too-many-branches
from enum import StrEnum, auto


class Bet(StrEnum):
    """An enumeration over the different bets."""

    WIN = auto()
    PLACE = auto()
    QUINELLA = auto()
    QUINELLA_PLACE = auto()
    FORECAST = auto()
    TIERCE = auto()
    TRIO = auto()
    FIRST_4 = auto()
    QUARTET = auto()
    FIRST_DOUBLE = auto()
    SECOND_DOUBLE = auto()
    FIRST_DOUBLE_TRIO = auto()
    THIRD_DOUBLE = auto()
    FOURTH_DOUBLE = auto()
    FIRST_TREBLE = auto()
    SECOND_DOUBLE_TRIO = auto()
    FIFTH_DOUBLE = auto()
    SIXTH_DOUBLE = auto()
    THIRD_DOUBLE_TRIO = auto()
    TRIPLE_TRIO = auto()
    TRIPLE_TRIO_CONSOLATION = auto()
    SEVENTH_DOUBLE = auto()
    EIGHTH_DOUBLE = auto()
    FOURTH_DOUBLE_TRIO = auto()
    NINTH_DOUBLE = auto()
    TENTH_DOUBLE = auto()
    SECOND_TREBLE = auto()
    FIFTH_DOUBLE_TRIO = auto()
    SIX_UP = auto()
    TREBLE = auto()


_BETS = {str(x).lower(): x for x in Bet}


def bet_from_str(bet_str: str) -> Bet:
    """Find a bet from a string."""
    bet_str = bet_str.lower()
    bet = _BETS.get(bet_str)
    if bet is None:
        if bet_str == "quinella place":
            return Bet.QUINELLA_PLACE
        if bet_str == "first 4":
            return Bet.FIRST_4
        if bet_str == "1st double":
            return Bet.FIRST_DOUBLE
        if bet_str == "2nd double":
            return Bet.SECOND_DOUBLE
        if bet_str == "1st double trio":
            return Bet.FIRST_DOUBLE_TRIO
        if bet_str == "3rd double":
            return Bet.THIRD_DOUBLE
        if bet_str == "4th double":
            return Bet.FOURTH_DOUBLE
        if bet_str == "1st treble":
            return Bet.FIRST_TREBLE
        if bet_str == "2nd double trio":
            return Bet.SECOND_DOUBLE_TRIO
        if bet_str == "5th double":
            return Bet.FIFTH_DOUBLE
        if bet_str == "6th double":
            return Bet.SIXTH_DOUBLE
        if bet_str == "3rd double trio":
            return Bet.THIRD_DOUBLE_TRIO
        if bet_str == "triple trio":
            return Bet.TRIPLE_TRIO
        if bet_str == "triple trio(consolation)":
            return Bet.TRIPLE_TRIO_CONSOLATION
        if bet_str == "7th double":
            return Bet.SEVENTH_DOUBLE
        if bet_str == "8th double":
            return Bet.EIGHTH_DOUBLE
        if bet_str == "4th double trio":
            return Bet.FOURTH_DOUBLE_TRIO
        if bet_str == "9th double":
            return Bet.NINTH_DOUBLE
        if bet_str == "10th double":
            return Bet.TENTH_DOUBLE
        if bet_str == "2nd treble":
            return Bet.SECOND_TREBLE
        if bet_str == "5th double trio":
            return Bet.FIFTH_DOUBLE_TRIO
        if bet_str == "six up":
            return Bet.SIX_UP
        raise ValueError(f"Unrecognised bet: {bet_str}")
    return bet



"""Aussportsbetting venue model."""

import datetime

import requests_cache

from ...cache import MEMORY
from ..google.google_address_model import create_google_address_model
from ..venue_model import VenueModel


@MEMORY.cache(ignore=["session"])
def create_aussportsbetting_venue_model(
    venue: str, session: requests_cache.CachedSession, dt: datetime.datetime
) -> VenueModel:
    """Create a venue model based off aus sports betting."""
    address_model = create_google_address_model(venue, session, dt)
    return VenueModel(
        identifier=venue,
        name=venue,
        address=address_model,
        is_grass=None,
        is_indoor=None,
        is_turf=None,
        is_dirt=None,
    )



"""HKJC HKJC coach model."""

import io

import pandas as pd
import pytest_is_running

from ....cache import MEMORY
from ....proxy_session import ProxySession
from ...coach_model import CoachModel


def _create_hkjc_hkjc_coach_model(
    session: ProxySession,
    url: str,
) -> CoachModel:
    with session.wayback_disabled():
        response = session.get(url)
    response.raise_for_status()

    handle = io.StringIO()
    handle.write(response.text)
    handle.seek(0)
    dfs = pd.read_html(handle)

    name = None
    age = None
    for count, df in enumerate(dfs):
        if count == 0:
            name = df.iat[0, 0].strip()
            age_str = df.iat[1, 0].strip().split(":")[-1].strip().split(",")[0].strip()
            if "–" in age_str:
                age_str = age_str.split("–")[-1].strip()
            if "-" in age_str:
                age_str = age_str.split("-")[-1].strip()
            age = int(age_str)

    if name is None:
        raise ValueError("name is null")

    return CoachModel(
        identifier=name,
        name=name,
        birth_date=None,
        age=age,
    )


@MEMORY.cache(ignore=["session"])
def _cached_create_hkjc_hkjc_coach_model(
    session: ProxySession,
    url: str,
) -> CoachModel:
    return _create_hkjc_hkjc_coach_model(session=session, url=url)


def create_hkjc_hkjc_coach_model(
    session: ProxySession,
    url: str,
) -> CoachModel:
    """Create a coach model based off HKJC."""
    if not pytest_is_running.is_running():
        return _cached_create_hkjc_hkjc_coach_model(session=session, url=url)
    return _create_hkjc_hkjc_coach_model(session=session, url=url)



"""Combined coach model."""

from ..coach_model import CoachModel
from .null_check import is_null


def create_combined_coach_model(
    coach_models: list[CoachModel], identifier: str
) -> CoachModel:
    """Create a coach model by combining many coach models."""
    name = None
    birth_date = None
    age = None
    for coach_model in coach_models:
        coach_model_name = coach_model.name
        if not is_null(coach_model_name):
            name = coach_model_name
        coach_model_birthdate = coach_model.birth_date
        if not is_null(coach_model_birthdate):
            birth_date = coach_model_birthdate
        coach_model_age = coach_model.age
        if not is_null(coach_model_age):
            age = coach_model_age
    if name is None:
        raise ValueError("name is null.")
    return CoachModel(
        identifier=identifier,
        name=name,
        birth_date=birth_date,
        age=age,
    )



"""NBA NBA.com league model."""

# pylint: disable=too-many-statements,protected-access,too-many-locals,bare-except,too-many-branches,duplicate-code
import datetime
import logging
from typing import Iterator

from ....proxy_session import ProxySession
from ...game_model import GameModel
from ...league import League
from ...league_model import LeagueModel
from .nba_nbacom_game_model import create_nba_nbacom_game_model


class NBANBAComLeagueModel(LeagueModel):
    """NBA NBA.com implementation of the league model."""

    def __init__(self, session: ProxySession, position: int | None = None) -> None:
        super().__init__(League.AFL, session, position=position)

    @classmethod
    def name(cls) -> str:
        return "nba-nbacom-league-model"

    @property
    def games(self) -> Iterator[GameModel]:
        with self.session.cache_disabled():
            date_slug = datetime.datetime.today().date().strftime("%Y%m%d")
            response = self.session.get(
                f"https://stats.nba.com/js/data/leaders/00_daily_lineups_{date_slug}.json"
            )
            if response.ok:
                lineup = response.json()
                for game in lineup["games"]:
                    yield create_nba_nbacom_game_model(game, self.session)
            else:
                logging.warning("Failed to fetch NBA daily lineups.")



"""A class for holding address information."""

from typing import Literal

from pydantic import BaseModel, Field

from .field_type import TYPE_KEY, FieldType
from .weather_model import WeatherModel

ADDRESS_LATITUDE_COLUMN: Literal["latitude"] = "latitude"
ADDRESS_LONGITUDE_COLUMN: Literal["longitude"] = "longitude"
ADDRESS_TIMEZONE_COLUMN: Literal["timezone"] = "timezone"


class AddressModel(BaseModel):
    """The class for representing an address."""

    city: str = Field(..., json_schema_extra={TYPE_KEY: FieldType.CATEGORICAL})
    state: str = Field(..., json_schema_extra={TYPE_KEY: FieldType.CATEGORICAL})
    zipcode: str | None = Field(
        ..., json_schema_extra={TYPE_KEY: FieldType.CATEGORICAL}
    )
    latitude: float | None = Field(..., alias=ADDRESS_LATITUDE_COLUMN)
    longitude: float | None = Field(..., alias=ADDRESS_LONGITUDE_COLUMN)
    housenumber: str | None = Field(
        ..., json_schema_extra={TYPE_KEY: FieldType.CATEGORICAL}
    )
    weather: WeatherModel | None
    timezone: str = Field(
        ...,
        json_schema_extra={TYPE_KEY: FieldType.CATEGORICAL},
        alias=ADDRESS_TIMEZONE_COLUMN,
    )
    country: str = Field(..., json_schema_extra={TYPE_KEY: FieldType.CATEGORICAL})



"""The enumeration of the different supported sexes."""

from enum import StrEnum


class Sex(StrEnum):
    """An enumeration over the different sexes."""

    MALE = "male"
    FEMALE = "female"
    GELDING = "gelding"
    STALLION = "stallion"
    FILLY = "filly"
    MARE = "mare"
    COLT = "colt"
    RIG = "rig"


_SEX = {str(x): x for x in Sex}


def sex_from_str(sex_str: str) -> Sex:
    """Find a sex from a string."""
    sex_str = sex_str.lower().strip()
    sex = _SEX.get(sex_str)
    if sex is None:
        if sex_str == "horse":
            return Sex.STALLION
        raise ValueError(f'Unrecognised sex: "{sex_str}"')
    return sex



"""NBA NBA.com venue model."""

# pylint: disable=too-many-statements,protected-access,duplicate-code
import datetime

import requests_cache

from ...google.google_address_model import create_google_address_model
from ...venue_model import VenueModel


def create_nba_nbacom_venue_model(
    venue_name: str,
    session: requests_cache.CachedSession,
    dt: datetime.datetime,
) -> VenueModel:
    """Create a game model from AFL Tables."""
    address_model = create_google_address_model(venue_name, session, dt)
    return VenueModel(
        identifier=venue_name,
        name=venue_name,
        address=address_model,
        is_grass=None,
        is_indoor=None,
        is_turf=None,
        is_dirt=None,
    )



"""AFL AFLTables coach model."""

# pylint: disable=too-many-locals
import datetime
import io
import logging
import os
import urllib.parse
from urllib.parse import urlparse

import pandas as pd
import pytest_is_running
import requests_cache
from bs4 import BeautifulSoup
from dateutil.parser import parse
from dateutil.relativedelta import relativedelta

from ....cache import MEMORY
from ...coach_model import CoachModel

_COACH_URL_CACHE: dict[str, str] = {}


def _create_afl_afltables_coach_model(
    url: str, session: requests_cache.CachedSession, year: int, dt: datetime.datetime
) -> CoachModel | None:
    html = _COACH_URL_CACHE.get(url)
    if html is None:
        with session.cache_disabled():
            response = session.get(url)
            html = response.text
            _COACH_URL_CACHE[url] = html

    handle = io.StringIO()
    handle.write(html)
    handle.seek(0)
    dfs = pd.read_html(handle)
    df = dfs[0]
    df = df[df.columns.values.tolist()[:2]]

    year_ranges = df[df.columns.values.tolist()[1]].tolist()

    names = df[df.columns.values.tolist()[0]].tolist()
    name = None
    for count, year_range in enumerate(year_ranges):
        if "-" in year_range:
            start_year, end_year = [int(x) for x in year_range.split("-")]
        else:
            try:
                start_year = end_year = int(year_range)
            except ValueError:
                continue
        if start_year <= year <= end_year:
            name = names[count].strip()
            break
    if name is None:
        return None

    coach_url = None
    soup = BeautifulSoup(html, "lxml")
    for a in soup.find_all("a", href=True):
        a_text = a.get_text().strip()
        if a_text == name:
            coach_url = urllib.parse.urljoin(url, a.get("href"))
            break
    if coach_url is None:
        raise ValueError("coach is null")
    o = urlparse(coach_url)
    last_component = o.path.split("/")[-1]
    identifier, _ = os.path.splitext(last_component)

    response = session.get(coach_url)
    response.raise_for_status()
    soup = BeautifulSoup(response.text, "lxml")

    birth_date = None
    try:
        birth_date = parse(soup.get_text().split("Born:")[1].strip().split()[0])
    except IndexError:
        logging.warning("Failed to find birth date from %s", response.url)

    return CoachModel(
        identifier=identifier,
        name=name,
        birth_date=birth_date,
        age=relativedelta(birth_date, dt).years if birth_date is not None else None,
    )


@MEMORY.cache(ignore=["session"])
def _cached_create_afl_afltables_coach_model(
    url: str, session: requests_cache.CachedSession, year: int, dt: datetime.datetime
) -> CoachModel | None:
    return _create_afl_afltables_coach_model(url=url, session=session, year=year, dt=dt)


def create_afl_afltables_coach_model(
    url: str, session: requests_cache.CachedSession, year: int, dt: datetime.datetime
) -> CoachModel | None:
    """Create a coach model from AFL tables."""
    if not pytest_is_running.is_running():
        return _cached_create_afl_afltables_coach_model(
            url=url, session=session, year=year, dt=dt
        )
    with session.cache_disabled():
        return _create_afl_afltables_coach_model(
            url=url, session=session, year=year, dt=dt
        )



"""The prototype class for a coach."""

import datetime
from typing import Literal

from pydantic import BaseModel, Field

from .field_type import TYPE_KEY, FieldType

COACH_IDENTIFIER_COLUMN: Literal["identifier"] = "identifier"
COACH_NAME_COLUMN: Literal["name"] = "name"
COACH_BIRTH_DATE_COLUMN: Literal["birth_date"] = "birth_date"
COACH_AGE_COLUMN: Literal["age"] = "age"


class CoachModel(BaseModel):
    """The serialisable coach class."""

    identifier: str = Field(
        ...,
        json_schema_extra={TYPE_KEY: FieldType.CATEGORICAL},
        alias=COACH_IDENTIFIER_COLUMN,
    )
    name: str = Field(..., alias=COACH_NAME_COLUMN)
    birth_date: datetime.date | None = Field(..., alias=COACH_BIRTH_DATE_COLUMN)
    age: int | None = Field(..., alias=COACH_AGE_COLUMN)



"""Sports Reference team model."""

# pylint: disable=too-many-arguments,too-many-locals,duplicate-code,too-many-branches,too-many-statements
import datetime
import http
import json
import logging
import urllib.parse
from urllib.parse import urlparse

import extruct  # type: ignore
import pytest_is_running
import requests
from bs4 import BeautifulSoup, Tag
from w3lib.html import get_base_url

from ...cache import MEMORY
from ...proxy_session import ProxySession
from ..google.google_news_model import create_google_news_models
from ..league import League
from ..sex import Sex
from ..team_model import TeamModel
from ..x.x_social_model import create_x_social_model
from .sportsreference_coach_model import create_sportsreference_coach_model
from .sportsreference_player_model import create_sportsreference_player_model

_BAD_URLS = {
    "https://www.sports-reference.com/cbb/players/jahmiah-simmons-2.html",
    "https://www.sports-reference.com/cbb/players/mohamed-sherif-12.html",
    "https://www.sports-reference.com/cbb/players/ana-beatriz-passos-alves-da-silva-1.html",
    "https://www.sports-reference.com/cbb/players/cia-eklof-1.html",
    "https://www.sports-reference.com/cbb/players/aj-caldwell-2.html",
    "https://www.sports-reference.com/cbb/players/akuwovo-ogheneyole-1.html",
    "https://www.sports-reference.com/cbb/players/jevon-lyle-1.html",
    "https://www.sports-reference.com/cbb/players/mike-aaman-1.html",
}
_NON_WAYBACK_URLS: set[str] = {
    "https://www.sports-reference.com/cbb/schools/stony-brook/women/2021.html",
    "https://www.sports-reference.com/cbb/schools/minnesota/women/2019.html",
    "https://www.sports-reference.com/cbb/schools/rice/women/2019.html",
    "https://www.sports-reference.com/cbb/schools/north-carolina/women/2018.html",
    "https://www.sports-reference.com/cbb/schools/minnesota/women/2009.html",
    "https://www.basketball-reference.com/teams/GSW/2016.html",
}
_BAD_TEAM_URLS = {
    "https://www.sports-reference.com/cbb/schools/mid-atlantic-christian/2016.html",
    "https://www.sports-reference.com/cbb/schools/claflin/2013.html",
    "https://www.sports-reference.com/cbb/schools/chaminade/2011.html",
}


def _find_name(response: requests.Response, soup: BeautifulSoup, url: str) -> str:
    base_url = get_base_url(response.text, url)
    try:
        data = extruct.extract(response.text, base_url=base_url)
        return data["json-ld"][0]["name"]
    except (json.decoder.JSONDecodeError, IndexError) as exc:
        h1 = soup.find("h1")
        if not isinstance(h1, Tag):
            logging.error(response.text)
            raise ValueError(f"h1 is null for {url}.") from exc
        span = h1.find_all("span")
        try:
            return span[1].get_text().strip()
        except IndexError:
            for span in soup.find_all("span", itemprop="title"):
                if not isinstance(span, Tag):
                    continue
                span_text = span.get_text().strip()
                if span_text == "BBR Home":
                    continue
                if span_text == "Teams":
                    continue
                return span_text
            test_url = "/".join(url.split("/")[:-1]) + "/"
            for a in soup.find_all("a", href=True):
                a_url = urllib.parse.urljoin(url, a.get("href"))
                if a_url == test_url:
                    return a.get_text().strip()
            name_tag = soup.find("meta", itemprop="name")
            if not isinstance(name_tag, Tag):
                name_tag = soup.find("meta", itemprop="og:title")
                if not isinstance(name_tag, Tag):
                    logging.error(type(name_tag))
                    logging.error(response.text)
                    raise ValueError("name_tag not a tag.") from exc
            content = name_tag.get("content")
            if not isinstance(content, str):
                raise ValueError("content not a tag.") from exc
            return " ".join(content.strip().split()[1:])


def _create_sportsreference_team_model(
    session: ProxySession,
    url: str,
    dt: datetime.datetime,
    league: League,
    player_urls: set[str],
    points: float,
    fg: dict[str, int],
    fga: dict[str, int],
    offensive_rebounds: dict[str, int],
    assists: dict[str, int],
    turnovers: dict[str, int],
    team_name: str,
    positions_validator: dict[str, str],
) -> TeamModel:
    if url in _BAD_TEAM_URLS:
        return TeamModel(
            identifier=team_name,
            name=team_name,
            location=None,
            players=[],
            odds=[],
            points=points,
            ladder_rank=None,
            news=create_google_news_models(team_name, session, dt, league),
            social=create_x_social_model(team_name, session, dt),
            coaches=[],
            lbw=None,
            end_dt=None,
        )

    if url in _NON_WAYBACK_URLS:
        with session.wayback_disabled():
            response = session.get(url)
    else:
        response = session.get(url)

    if response.status_code == http.HTTPStatus.NOT_FOUND:
        logging.warning("Could not find team %s at url %s", team_name, url)
        return TeamModel(
            identifier=team_name,
            name=team_name,
            location=None,
            players=[],
            odds=[],
            points=points,
            ladder_rank=None,
            news=create_google_news_models(team_name, session, dt, league),
            social=create_x_social_model(team_name, session, dt),
            coaches=[],
            lbw=None,
            end_dt=None,
        )
    response.raise_for_status()

    soup = BeautifulSoup(response.text, "lxml")
    title = soup.find("title")
    if not isinstance(title, Tag):
        raise ValueError(f"title not a tag for {url}.")
    title_str = title.get_text().strip().lower()
    if "file not found" in title_str:
        session.cache.delete(urls=[url])
        response = session.get(url)
        response.raise_for_status()

    name = _find_name(response, soup, url)

    valid_player_urls = set()
    for a in soup.find_all("a"):
        player_url = urllib.parse.urljoin(url, a.get("href"))
        if player_url in player_urls and player_url not in _BAD_URLS:
            valid_player_urls.add(player_url)

    coach_url = None
    for a in soup.find_all("a", href=True):
        a_url = urllib.parse.urljoin(url, a.get("href"))
        a_o = urlparse(a_url)
        path_split = a_o.path.split("/")
        if len(path_split) <= 2:
            continue
        if not path_split[-1]:
            continue
        entity_identifier = path_split[-2]
        if entity_identifier == "coaches":
            coach_url = a_url
            break

    o = urlparse(url)
    sex_id = o.path.split("/")[-2]
    sex = Sex.MALE
    if sex_id == "women":
        sex = Sex.FEMALE

    positions = {}
    for tr in soup.find_all("tr"):
        player_name = None
        position = None
        for td in tr.find_all("td"):
            data_stat = td.get("pos")
            if data_stat == "player":
                player_name = td.get_text().strip()
            elif data_stat == "pos":
                position = td.get_text().strip()
        if player_name is not None and position is not None:
            positions[player_name] = position

    return TeamModel(
        identifier=name,
        name=name,
        players=[
            y
            for y in [  # pyright: ignore
                create_sportsreference_player_model(
                    session=session,
                    player_url=x,
                    fg=fg,
                    fga=fga,
                    offensive_rebounds=offensive_rebounds,
                    assists=assists,
                    turnovers=turnovers,
                    positions=positions,
                    positions_validator=positions_validator,
                    sex=sex,
                    dt=dt,
                )
                for x in valid_player_urls
            ]
            if y is not None
        ],
        odds=[],
        points=points,
        ladder_rank=None,
        location=None,
        news=create_google_news_models(name, session, dt, league),
        social=create_x_social_model(name, session, dt),
        coaches=[create_sportsreference_coach_model(session, coach_url, dt)]
        if coach_url is not None
        else [],
        lbw=None,
        end_dt=None,
    )


@MEMORY.cache(ignore=["session"])
def _cached_create_sportsreference_team_model(
    session: ProxySession,
    url: str,
    dt: datetime.datetime,
    league: League,
    player_urls: set[str],
    points: float,
    fg: dict[str, int],
    fga: dict[str, int],
    offensive_rebounds: dict[str, int],
    assists: dict[str, int],
    turnovers: dict[str, int],
    team_name: str,
    positions_validator: dict[str, str],
) -> TeamModel:
    return _create_sportsreference_team_model(
        session,
        url,
        dt,
        league,
        player_urls,
        points,
        fg,
        fga,
        offensive_rebounds,
        assists,
        turnovers,
        team_name,
        positions_validator,
    )


def create_sportsreference_team_model(
    session: ProxySession,
    url: str,
    dt: datetime.datetime,
    league: League,
    player_urls: set[str],
    points: float,
    fg: dict[str, int],
    fga: dict[str, int],
    offensive_rebounds: dict[str, int],
    assists: dict[str, int],
    turnovers: dict[str, int],
    team_name: str,
    positions_validator: dict[str, str],
) -> TeamModel:
    """Create a team model from Sports Reference."""
    if not pytest_is_running.is_running():
        return _cached_create_sportsreference_team_model(
            session,
            url,
            dt,
            league,
            player_urls,
            points,
            fg,
            fga,
            offensive_rebounds,
            assists,
            turnovers,
            team_name,
            positions_validator,
        )
    with session.cache_disabled():
        return _create_sportsreference_team_model(
            session,
            url,
            dt,
            league,
            player_urls,
            points,
            fg,
            fga,
            offensive_rebounds,
            assists,
            turnovers,
            team_name,
            positions_validator,
        )



# pylint: disable=missing-module-docstring,missing-class-docstring,invalid-name,bare-except,consider-using-f-string,unused-variable,broad-exception-raised,inconsistent-return-statements,raise-missing-from,no-else-return,too-many-arguments,singleton-comparison
# ruff: noqa: E722,E712
# type: ignore
import urllib

import feedparser  # type: ignore
import requests
from bs4 import BeautifulSoup
from dateparser import parse as parse_date


class GoogleNews:
    def __init__(self, lang="en", country="US", session=None):
        self.lang = lang.lower()
        self.country = country.upper()
        self.BASE_URL = "https://news.google.com/rss"
        self._session = session if session is not None else requests.Session()
        self.timeout = 30.0

    def __top_news_parser(self, text):
        """Return subarticles from the main and topic feeds"""
        try:
            bs4_html = BeautifulSoup(text, "lxml")
            # find all li tags
            lis = bs4_html.find_all("li")
            sub_articles = []
            for li in lis:
                try:
                    sub_articles.append(
                        {
                            "url": li.a["href"],
                            "title": li.a.text,
                            "publisher": li.font.text,
                        }
                    )
                except:
                    pass
            return sub_articles
        except:
            return text

    def __ceid(self):
        """Compile correct country-lang parameters for Google News RSS URL"""
        return "?ceid={}:{}&hl={}&gl={}".format(
            self.country, self.lang, self.lang, self.country
        )

    def __add_sub_articles(self, entries):
        for i, val in enumerate(entries):
            if "summary" in entries[i].keys():
                entries[i]["sub_articles"] = self.__top_news_parser(
                    entries[i]["summary"]
                )
            else:
                entries[i]["sub_articles"] = None
        return entries

    def __scaping_bee_request(self, api_key, url):
        response = self._session.get(
            url="https://app.scrapingbee.com/api/v1/",
            params={"api_key": api_key, "url": url, "render_js": "false"},
            timeout=self.timeout,
        )
        if response.status_code == 200:
            return response
        if response.status_code != 200:
            raise Exception(
                "ScrapingBee status_code: "
                + str(response.status_code)
                + " "
                + response.text
            )

    def __parse_feed(self, feed_url, proxies=None, scraping_bee=None):
        if scraping_bee and proxies:
            raise Exception("Pick either ScrapingBee or proxies. Not both!")

        if proxies:
            r = self._session.get(feed_url, proxies=proxies, timeout=self.timeout)
        elif scraping_bee:
            r = self.__scaping_bee_request(url=feed_url, api_key=scraping_bee)
        else:
            r = self._session.get(feed_url, timeout=self.timeout)

        if "https://news.google.com/rss/unsupported" in r.url:  # pyright: ignore
            raise Exception("This feed is not available")

        d = feedparser.parse(r.text)

        return dict((k, d[k]) for k in ("feed", "entries"))

    def __search_helper(self, query):
        return urllib.parse.quote_plus(query)

    def __from_to_helper(self, validate=None):
        try:
            validate = parse_date(validate).strftime("%Y-%m-%d")
            return str(validate)
        except:
            raise Exception("Could not parse your date")

    def top_news(self, proxies=None, scraping_bee=None):
        """Return a list of all articles from the main page of Google News
        given a country and a language"""
        d = self.__parse_feed(
            self.BASE_URL + self.__ceid(), proxies=proxies, scraping_bee=scraping_bee
        )
        d["entries"] = self.__add_sub_articles(d["entries"])
        return d

    def topic_headlines(self, topic: str, proxies=None, scraping_bee=None):
        """Return a list of all articles from the topic page of Google News
        given a country and a language"""
        # topic = topic.upper()
        if topic.upper() in [
            "WORLD",
            "NATION",
            "BUSINESS",
            "TECHNOLOGY",
            "ENTERTAINMENT",
            "SCIENCE",
            "SPORTS",
            "HEALTH",
        ]:
            d = self.__parse_feed(
                self.BASE_URL
                + "/headlines/section/topic/{}".format(topic.upper())
                + self.__ceid(),
                proxies=proxies,
                scraping_bee=scraping_bee,
            )

        else:
            d = self.__parse_feed(
                self.BASE_URL + "/topics/{}".format(topic) + self.__ceid(),
                proxies=proxies,
                scraping_bee=scraping_bee,
            )

        d["entries"] = self.__add_sub_articles(d["entries"])
        if len(d["entries"]) > 0:
            return d
        else:
            raise Exception("unsupported topic")

    def geo_headlines(self, geo: str, proxies=None, scraping_bee=None):
        """Return a list of all articles about a specific geolocation
        given a country and a language"""
        d = self.__parse_feed(
            self.BASE_URL + "/headlines/section/geo/{}".format(geo) + self.__ceid(),
            proxies=proxies,
            scraping_bee=scraping_bee,
        )

        d["entries"] = self.__add_sub_articles(d["entries"])
        return d

    def search(
        self,
        query: str,
        helper=True,
        when=None,
        from_=None,
        to_=None,
        proxies=None,
        scraping_bee=None,
    ):
        """
        Return a list of all articles given a full-text search parameter,
        a country and a language

        :param bool helper: When True helps with URL quoting
        :param str when: Sets a time range for the artiles that can be found
        """

        if when:
            query += " when:" + when

        if from_ and not when:
            from_ = self.__from_to_helper(validate=from_)
            query += " after:" + from_

        if to_ and not when:
            to_ = self.__from_to_helper(validate=to_)
            query += " before:" + to_

        if helper == True:
            query = self.__search_helper(query)

        search_ceid = self.__ceid()
        search_ceid = search_ceid.replace("?", "&")

        d = self.__parse_feed(
            self.BASE_URL + "/search?q={}".format(query) + search_ceid,
            proxies=proxies,
            scraping_bee=scraping_bee,
        )

        d["entries"] = self.__add_sub_articles(d["entries"])
        return d
